
//------------------------------------------------------------------------------
// module:   canDriver.cpp
// created:  P.K. Bartkiewicz, 27.11.2006
// modification history:
//
//------------------------------------------------------------------------------

#include <errno.h>
#include <sys/ioctl.h>
#include "canDriver.h"

CanDriver::CanDriver()
{
  fdCAN = -1;
  sprintf( device, "/dev/pcan%d", PCAN_USB );  // defualt USB interface
  setName( "CanDriver" );
}

CanDriver::~CanDriver()
{
  closeCAN();
}

CanDriver::CanDriver( unsigned char interfaceCode, unsigned int baudRate )
{ 
  fdCAN = -1;
  sprintf( device, "/dev/pcan%d", interfaceCode );
  initialize( baudRate );
  setName( "CanDriver" );
}

int CanDriver::initialize(  unsigned int baudRate )
{  
  openCAN( baudRate );
  if( Error ) Lg << ERROR << "--Error = " << Error << endl;
  return Error;
}

void CanDriver::openCAN( unsigned int baudRate  )
{
    TPCANInit init;
    TPBTR0BTR1 ratix;
     
    if( fdCAN > -1 )  closeCAN();
    if( ( fdCAN = open( device, O_RDWR )) < 0 )
    {
      Lg << ERROR << "--Error: can not open CAN device: " << device  << ", error code = " <<  errno  << endl;
      Lg << ERROR << "--Error description: " <<  strerror( errno ) << endl;
      exit(errno);
    }
    
    speed = baudRate;  
    ratix.dwBitRate = speed;
    ratix.wBTR0BTR1 = 0;
    Lg << DBG << "CAN: SETTING SPEED " << speed << " bit/s" << endl;
    Error = ioctl( fdCAN, PCAN_BTR0BTR1, &ratix);
    if( Error ) Lg << ERROR  <<  "--Error: setting speed error code: "<<  Error << endl;
    if (Error) return;
    
    init.wBTR0BTR1    = ratix.wBTR0BTR1;    // combined BTR0 and BTR1 register of the SJA100
    init.ucCANMsgType =  MSGTYPE_STANDARD;  // 11 or 29 bits
    init.ucListenOnly = 0;                  // listen only mode when != 0
    
    Error = ioctl(fdCAN, PCAN_INIT, &init); 
}

void CanDriver::closeCAN()
{
    close( fdCAN );
    Lg << DBG << "CAN interface closed"  << endl;
}

int  CanDriver::sendMsg( CanMessage&  m, long timeout )
{
    TPCANMsg msgBuff;
    fd_set fdset;
    struct timeval tout;
    
    FD_ZERO( &fdset );
    FD_SET( fdCAN, &fdset );

    tout.tv_sec = timeout / 1000;
    tout.tv_usec  = (timeout % 1000) * 1000;   /* convert to uSeconds */
    if( select( fdCAN + 1, NULL, &fdset, NULL, &tout ) > 0 )
    {	
	msgBuff.ID = m.getId();
	msgBuff.MSGTYPE = (m.isRTR() ? MSGTYPE_RTR : MSGTYPE_STANDARD) ;
	msgBuff.LEN = m.getDataLen();
	memcpy( msgBuff.DATA, m.getData(), MSG_DATA_LEN );       
	Error = ioctl( fdCAN, PCAN_WRITE_MSG, &msgBuff );
	if(Error) Lg << ERROR << "--Error: sending msg error code: " <<  Error << endl;
	return Error;
    }
    Lg << WARNING << "-Warning,  CanDriver:: transmitter full" << endl;
    return 0;
}


int CanDriver::receiveMsg(CanMessage& m, long timeout )
{
    fd_set fdset;
    struct timeval tout;   
    
    FD_ZERO( &fdset );
    FD_SET( fdCAN, &fdset );    
    
    tout.tv_sec = timeout / 1000;
    tout.tv_usec  = (timeout % 1000) * 1000;   /* convert to uSeconds */
    if( select( fdCAN + 1, &fdset, NULL, NULL, &tout ) > 0 )  return readMsg( m );
    return 0;  
}

int CanDriver::readMsg( CanMessage& m )
{
    TPCANRdMsg msgBuff;
    if( (Error = ioctl(fdCAN, PCAN_READ_MSG, &msgBuff)) < 0) return Error;
  
    if( msgBuff.Msg.MSGTYPE == MSGTYPE_RTR ) m.setRTR();
    m.setId( msgBuff.Msg.ID );
    m.setDataLen(msgBuff.Msg.LEN);
    memcpy( m.getData(), msgBuff.Msg.DATA, MSG_DATA_LEN );
   
    return 1;
}

int CanDriver::send( CanMessage &m )
{
  return sendQueue.put( m ) ? 0 : -1;
}


int CanDriver::receive( CanMessage &m )
{
  return receiveQueue.get( m );
}

void CanDriver::activate( void )
{
  // check, if any messages are pending in the CAN driver queue
  // if yes, read all of them and put into output queue
    
  CanMessage m;
      
  while( receiveQueue.available() )
    {
      if( !receiveMsg( m ) ) break;
      if( !receiveQueue.put( m ) ) Lg << ERROR << "--Error: receiver queue full, message(s) lost" << endl;      
    }

    while( sendQueue.getNpending() != 0 )
    {
      if( sendQueue.get( m ) ) sendMsg( m );
    }

    return;
}

