

#include "canDriverEsd.h"

extern Log Lg;

const ERRORDESCR CanDriverEsd::errorDescriptor[20] =
{
    { NTCAN_SUCCESS, "no error" },
    { NTCAN_RX_TIMEOUT, "rx tiemout" },
    { NTCAN_TX_TIMEOUT, "tx timeout" },
    { NTCAN_TX_ERROR, " tx error" },
    { NTCAN_CONTR_OFF_BUS, "bus off"},          
    { NTCAN_CONTR_BUSY, "controller busy" },
    { NTCAN_CONTR_WARN, "cpntroller warning" },
    { NTCAN_NO_ID_ENABLED, "no ID enabled"  },
    { NTCAN_ID_ALREADY_ENABLED, "ID already enabled" },
    { NTCAN_ID_NOT_ENABLED, "ID not enabled" },
    { NTCAN_INVALID_FIRMWARE, "invalid firmware" },
    { NTCAN_MESSAGE_LOST, "message lost" },
    { NTCAN_INVALID_HARDWARE, "invalid hardware" },
    { NTCAN_PENDING_WRITE, "write pending" },
    { NTCAN_PENDING_READ, "read pending" },
    { NTCAN_INVALID_DRIVER, "invalid driver" },
    { NTCAN_WRONG_DEVICE_STATE, "wrong device state" },
    { NTCAN_SOCK_CONN_TIMEOUT, "socket connecton timeout" },
    { NTCAN_SOCK_CMD_TIMEOUT, "socket command timeout" },
    { NTCAN_SOCK_HOST_NOT_FOUND, "socket host not found" }
};

const long CanDriverEsd::SupportedBaudRates[14] = 
                                           { 
                                            1000000, 
                                            666600, 
                                            500000, 
                                            333300, 
                                            250000, 
                                            166000,
                                            125000,
                                            100000,
                                             66600,
                                             50000,
                                             33300,
                                             20000,
                                             12500,
                                             10000 
                                           };
                                           
CanDriverEsd::CanDriverEsd()
{
    setName( "CanDriverEsd" );
}

CanDriverEsd::CanDriverEsd( long baudRate, unsigned int net )
{
    setName( "CanDriverEsd" );
    initialize( baudRate, net );
}

int CanDriverEsd::initialize(  long baudRate, unsigned int net )
{  
    int bdRateIdx = findBaudRateIndex( baudRate );
    if( bdRateIdx < 0 ) 
    {
        Lg << FATAL << "---Fatal: CanDriverEsd:: baud rate = " << baudRate << " is not supported." << endl;
        exit(1);
    }  
  int retCode = canOpen(    net,
                            0,   // mode, 
                            1024, // txQueueSize, 
                            1024, // rxQueueSize, 
                            10,  // txTimeOut, ms
                            1, // rxTimeOut, ms
                            &handle );        
    if( retCode != NTCAN_SUCCESS ) 
    {
        Lg << FATAL << "---Fatal: CanDriverEsd::initialize, open(): retCode = " <<  retCode  << ", " << getErrorDescription(retCode) << endl;
        exit(1);
    }             
    retCode = canSetBaudrate( handle, bdRateIdx);    
    if( retCode != NTCAN_SUCCESS )
    {
        Lg << FATAL << "---Fatal: CanDriverEsd::initialize, canSetBaudrate(): retCode = " <<  retCode << ", " << getErrorDescription(retCode) << endl;
        canClose( handle );
        exit(1);
    }
    
    // adding ids (filtering):    
    for(int i = 0; i < 2048; i++ )
    {
        retCode = canIdAdd( handle, i );    
        if( retCode != NTCAN_SUCCESS )
        { 
            Lg << FATAL <<  "---Fatal: CanDriverEsd::initialize, canIdAdd, retCode = " << retCode << ", " << getErrorDescription(retCode) << endl;
            canClose( handle );
            exit(1);
        }
    }  
   
  return 0;
}

int  CanDriverEsd::sendMsg( CanMessage&  m )
{
    CMSG msg;
    long len = 1;
    
    msg.id = m.getId(); 
    msg.len = m.getDataLen() | ( m.isRTR() ? 0x10 : 0x00 );
    memcpy( msg.data, m.getData(), 8 );    
  
    int errCode = canWrite( handle, &msg, &len, NULL ); // blocking write()
    //int errCode = canSend( handle, &msg, &len );
    if( errCode != NTCAN_SUCCESS )  Lg << ERROR << "--Error: sendMsg error code = 0x" << hex << errCode << dec << ", " << getErrorDescription(errCode) << endl;
    else if( len < 1 ) Lg << ERROR << "--Error: sending failed, but no error code" << endl;
    return errCode == NTCAN_SUCCESS;
}


int CanDriverEsd::receiveMsg( CanMessage& m )
{
    long len = 1;
    CMSG msg;
   
    // int errCode = canRead(handle,&msg, &len, NULL);  // blocking read()
    int errCode = canTake(handle,&msg, &len);
    if( errCode == NTCAN_SUCCESS )
    {
        if( len == 1 )
        {
            m.setId( msg.id );
            if( msg.len & 0x10 ) m.setRTR();
            m.setDataLen( msg.len & 0x0F );  // remove possible RTR flag
            m.setData( (char*)msg.data, m.getDataLen() );
            return 1;
        }
        return 0;
    }
    if (errCode == NTCAN_RX_TIMEOUT ) return 0;
    Lg << ERROR << "--Error: receiveMsg error code = 0x" << hex << errCode << dec  << ", " << getErrorDescription(errCode) << endl;
    return 0;  // error
}

int CanDriverEsd::findBaudRateIndex( long baudRate )
{
    for(unsigned int i = 0; i < sizeof( SupportedBaudRates ) / sizeof( long ); i++ )
    {
        if( baudRate == SupportedBaudRates[i] ) return i;
    }
    return -1;
}

const char * CanDriverEsd::getErrorDescription( int error )
{
    for( unsigned int i = 0; i < sizeof( errorDescriptor )/ sizeof( errorDescriptor[0]); i++ )
    {
        if( errorDescriptor[i].code == error ) return errorDescriptor[i].message;
    }
    return "unknown ESD interface error";
}
