
#include "canNodeImage.h"
#include "flashProgrammer.h"
#include "canMessageDispatcher.h"
#include <tine/R4/tinetype.h>

#define __PRINT_FUNCTION_NAME__     /* Lg << DBG << "Inside: " << (char*) __FUNCTION__  << " for node: 0x" << hex << getNodeId() << dec << endl;  */

CanNodeImage::CanNodeImage(CanMessageDispatcher * cmd, CanMessageCollector * collector, char * dcfFile, char * nodeName, int nodeId)
{
 
  construct(cmd);
  connectMessageCollector(collector);
  
  parse( dcfFile, nodeName, nodeId );
  
  // register CAN message handlers
  cmd->addMessageHandler((short)(0x700 + getNodeId()), this, &CanNodeImage::nmtHandler, 0 );
  cmd->addMessageHandler((short)(0x580 + getNodeId()), this, &CanNodeImage::sdoHandler, 0 );
  cmd->addMessageHandler((short)(0x80 + getNodeId()), this, &CanNodeImage::emcyHandler, 0 );

  //TINE interface
  buildPropertiesSet();
 
}

CanNodeImage::CanNodeImage(CanMessageDispatcher * cmd, CanMessageCollector * collector, char * dcfFile)
{
 
  construct(cmd);
  connectMessageCollector(collector);  
  if( dcfFile )parse( dcfFile );

  // register CAN message handlers
  cmd->addMessageHandler((short)(0x700 + getNodeId()), this, &CanNodeImage::nmtHandler, 0 );
  cmd->addMessageHandler((short)(0x580 + getNodeId()), this, &CanNodeImage::sdoHandler, 0 );
  cmd->addMessageHandler((short)(0x80 + getNodeId()), this, &CanNodeImage::emcyHandler, 0 );

  //TINE interface
  buildPropertiesSet();
 
}
void CanNodeImage::construct(CanMessageDispatcher * cmd)
{
  ready = 0;
  ignoreBootup = 0;
  ignoreBootupTimer = new HrTimer();
  memset( &canNodeIdInfo, '\0', sizeof(CanNodeIdInfo) ); 
  memset( TPDOvariables, '\0', sizeof(TPDOvariables) );
  memset( RPDOvariables, '\0', sizeof(RPDOvariables) );

  EdsParser = new EDSparser();
  od = new ObjectDictionary();
  sdoClient = new SDOclient();
  sdoForwarder = new SDOforwarder(sdoClient);
  pdoClient = new PDOclient(this);
  nodeGuardProducer = new NodeGuardProducer();
  heartBeatConsumer = new HeartBeatConsumer( this, 1000 );
  sequencer = new Sequencer( this );
  flashProgrammer = new FlashProgrammer(this);

  // add devices to the polling list
  add( sdoClient ); 
  add( sdoForwarder );
  add( pdoClient );  
  add( nodeGuardProducer );
  add( heartBeatConsumer );
  add( sequencer );
  add( flashProgrammer );
  canMessageCollector = NULL;
  canMessageDispatcher = cmd;
  
  setName( "CanNode" );
  setFamilyName( "CanNodeImage" );
  setInitializationError( INIT_NO_ERROR );
  nodeId = 0; 
  server = NULL; 
  needInitialization = 0;
  nodeGuardNeeded = 0;
}

CanNodeImage::~CanNodeImage()
{
  delete sequencer;
  delete heartBeatConsumer; 
  delete nodeGuardProducer; 
  delete pdoClient; 
  delete sdoClient;
  delete od;
  delete EdsParser;
}
 
void CanNodeImage::buildPropertiesSet()
{
  removeProperties();  // make the property list empty

  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, nodeId), 1, CF_LONG ); 
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, deviceType),1,CF_LONG); 
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, vendorId),1,CF_LONG);
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, productCode),1,CF_LONG);
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, revisionNumber),1,CF_LONG);
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, serialNumber),1,CF_LONG);
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, manufactureDeviceName), SHORT_STRING_LEN+1, CF_BYTE); 
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, hardwareVersion), SHORT_STRING_LEN+1, CF_BYTE);
  fillstruct("NODEIDINFO",OFFSETIN(CanNodeIdInfo, softwareVersion), SHORT_STRING_LEN+1, CF_BYTE);
  fillstruct("NODEIDINFO",sizeof(CanNodeIdInfo),1,CF_NULL);
/*
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, index), 1, CF_INTEGER);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, subindex), 1, CF_INTEGER);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, dataValue), 1, CF_LNGINT);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, updateTime), 1, CF_LNGINT);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, dataSize), 1, CF_INTEGER);  
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, dataFormat), 1, CF_INTEGER);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, npdos), 1, CF_INTEGER);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, pdos), MAX_PDO, CF_BYTE);
  fillstruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, variableName), 64, CF_BYTE);
  fillstruct( "MAPPEDVAR", sizeof(MappedVariableValue),1,CF_NULL);
  */
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, index), 1, CF_INTEGER, "index");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, subindex), 1, CF_INTEGER, "subindex");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, dataValue), 1, CF_LNGINT, "dataValue");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, updateTime), 1, CF_LNGINT, "updateTime");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, dataSize), 1, CF_INTEGER, "dataSize");  
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, dataFormat), 1, CF_INTEGER, "dataFormat");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, npdos), 1, CF_INTEGER, "nPdos");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, pdos), MAX_PDO, CF_INTEGER, "PDOs");
  AddFieldToStruct( "MAPPEDVAR", OFFSETIN(MappedVariableValue, variableName), 64, CF_BYTE, "Variablename");
  SealTaggedStruct( "MAPPEDVAR", sizeof(MappedVariableValue), MAX_PDO * 8);
  
  
  fillstruct( "ODENTRYINFO", OFFSETIN(ODentryDataInfo, index), 1 , CF_LONG );
  fillstruct( "ODENTRYINFO", OFFSETIN(ODentryDataInfo, subindex), 1 , CF_LONG );  
  fillstruct( "ODENTRYINFO", OFFSETIN(ODentryDataInfo, dataCanType), 1 , CF_LONG );
  fillstruct( "ODENTRYINFO", OFFSETIN(ODentryDataInfo, dataSize), 1 , CF_LONG );
  fillstruct( "ODENTRYINFO", OFFSETIN(ODentryDataInfo, data), 512 , CF_BYTE );
  fillstruct( "ODENTRYINFO", sizeof(ODentryDataInfo),1,CF_NULL);
  
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, timeStamp), 2, CF_LONG );
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, emcyCode), 1, CF_SHORT );
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, errorRegister), 1, CF_BYTE );
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, manufactureSpecificCode), 5, CF_BYTE );
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, isMessage), 1, CF_BYTE );
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, errorRegisterText), 207, CF_BYTE );
  fillstruct( "EMCY", OFFSETIN(EMCY_INFO, emcyText), 48, CF_BYTE );
  fillstruct( "EMCY", sizeof(EMCY_INFO),1,CF_NULL);
    
  DTYPE out, in;
  out.dArrayLength = 1;  out.dFormat = CF_STRUCT; strcpy( out.dTag, "NODEIDINFO" );
  in.dArrayLength = 0;  in.dFormat = CF_NULL; 
  addProperty( "IdInfo", &out, &in, CA_READ, "Device ID", &CanNodeImage::property_IdInfo);  
  
  out.dArrayLength = 1;  out.dFormat = CF_LONG; 
  in.dArrayLength = 1;  in.dFormat = CF_LONG; 
  addProperty("NMState",&out,&in, CA_READ | CA_WRITE, "Network Management State",  &CanNodeImage::property_NMState);
  
  out.dArrayLength = 1;  out.dFormat = CF_LONG; 
  in.dArrayLength = 1;  in.dFormat = CF_LONG; 
  addProperty("NGuard_onOff",&out, &in, CA_READ | CA_WRITE, "Node Guard On/Off", &CanNodeImage::property_nodeGuardState );  
  
  out.dArrayLength = 1;  out.dFormat = CF_LONG; 
  in.dArrayLength = 1;  in.dFormat = CF_LONG; 
  addProperty("NGuard_period", &out, &in, CA_READ | CA_WRITE, "Node Guard period", &CanNodeImage::property_nodeGuardPeriod);
  
  out.dArrayLength = 1;  out.dFormat = CF_STRUCT; strcpy( out.dTag, "ODENTRYINFO" );
  in.dArrayLength = 1;  in.dFormat = CF_STRUCT; strcpy( in.dTag, "ODENTRYINFO" );
  addProperty("ODentry",&out, &in, CA_READ | CA_WRITE, "Obj Dictionary entry read/write", &CanNodeImage::property_odEntry);
  
  out.dArrayLength = 1;  out.dFormat = CF_LONG; strcpy( out.dTag, "" );
  in.dArrayLength = 3;  in.dFormat = CF_LONG; strcpy( in.dTag, "" );
  addProperty("ODentry",&out, &in, CA_READ | CA_WRITE, "Obj Dictionary: idx, subidx, data", &CanNodeImage::property_odEntry);
  
  out.dArrayLength = 512;  out.dFormat = CF_TEXT; strcpy( out.dTag, "" );
  in.dArrayLength = 520;  in.dFormat = CF_LONG; strcpy( in.dTag, "" );
  addProperty("ODentry",&out, &in, CA_READ | CA_WRITE, "Obj Dictionary: idx, subidx, data", &CanNodeImage::property_odEntry);
  
  out.dArrayLength = 10;  out.dFormat = CF_LONG; 
  in.dArrayLength = 0;  in.dFormat = CF_NULL;
  addProperty("StateStat",&out, &in, CA_READ, "NMT states statistics", &CanNodeImage::property_stateStat );
  
  out.dArrayLength = 2;  out.dFormat = CF_LONG; 
  in.dArrayLength = 2;  in.dFormat = CF_LONG;
  addProperty("HbeatControl", &out, &in, CA_READ | CA_WRITE, "Heartbeat/node guard control", &CanNodeImage::property_hbeatControl);
  
  out.dArrayLength = MAX_PDO * 8;  out.dFormat = CF_STRUCT; strcpy( out.dTag, "MAPPEDVAR" );
  in.dArrayLength = 0;  in.dFormat = CF_NULL;
  addProperty("TPDOvariables", &out, &in, CA_READ, "List of process variables (TPDO)", &CanNodeImage::property_tpdoVariables); 
  addProperty("RPDOvariables", &out, &in, CA_READ, "List of process variables (RPDO)", &CanNodeImage::property_rpdoVariables);
  
  out.dArrayLength = 1;  out.dFormat = CF_STRUCT; strcpy( out.dTag, "EMCY" );
  in.dArrayLength = 0;  in.dFormat = CF_NULL;
  addProperty("EMCY", &out, &in, CA_READ, "Emergency message", &CanNodeImage::property_emcy);

  out.dArrayLength = 1;  out.dFormat = CF_INT32;
  in.dArrayLength = 256;  in.dFormat = CF_TEXT;
  addProperty("REPROGRAM", &out, &in, CA_WRITE, "Firmware upgrade", &CanNodeImage::property_firmwareUpgrade);

  out.dArrayLength = 4;  out.dFormat = CF_INT32;
  in.dArrayLength = 0;  in.dFormat = CF_NULL;
  addProperty("REPROGRAM.STATUS", &out, &in, CA_READ, "Firmware upgrade status", &CanNodeImage::property_firmwareUpgradeStatus);

  
  out.dArrayLength = 0;  out.dFormat = CF_NULL;
  in.dArrayLength = 1;  in.dFormat = CF_INT32;
  addProperty("REPROGRAM.ERASEANDPROG", &out, &in, CA_WRITE, "Erase and reprogram flash", &CanNodeImage::property_firmwareUpgradeEraseAndReprogram);

  out.dArrayLength = 0;  out.dFormat = CF_NULL;
  in.dArrayLength = 256;  in.dFormat = CF_TEXT;
  addProperty("REPROGRAM.SENDDATA", &out, &in, CA_WRITE, "Send data", &CanNodeImage::property_firmwareUpgradeSendData);
  
  out.dArrayLength = 0;  out.dFormat = CF_NULL;
  in.dArrayLength = 1;  in.dFormat = CF_INT32;
  addProperty("REPROGRAM.SWITCHTOMAIN", &out, &in, CA_WRITE, "Switch to main program", &CanNodeImage::property_firmwareUpgradeSwitchToMainProgram);
  
  out.dArrayLength = 0;  out.dFormat = CF_NULL;
  in.dArrayLength = 1;  in.dFormat = CF_INT32;
  addProperty("REPROGRAM.SWITCHTOPROGRAM", &out, &in, CA_WRITE, "Switch to programming mode", &CanNodeImage::property_firmwareUpgradeSwitchToProgrammingMode);
  
  out.dArrayLength = 2;  out.dFormat = CF_INT32;
  in.dArrayLength = 1;  in.dFormat = CF_INT32;
  addProperty("REPROGRAM.SUMS", &out, &in, CA_WRITE, "Calculate control sums", &CanNodeImage::property_firmwareUpgradeGetControlSums);

  int i,n;
  MappedVariable * mv;
  ODentry * ode;
  Lg << DBG << "Node = 0x" << hex << nodeId << dec << ": TPDO automatic variables: " << endl;
  n = pdoClient->getTpdoVariablesCount(); 
  for( i = 0; i < n; i++ )
    { 
      mv = pdoClient->getTmappedVariable(i);

      TPDOvariables[i].index = mv->index;
      TPDOvariables[i].subindex = mv->subindex;
      TPDOvariables[i].dataValue = 0; //
      TPDOvariables[i].updateTime = 0; //
      TPDOvariables[i].dataSize = mv->dataSize;
      TPDOvariables[i].npdos = mv->npdos;
      memcpy( TPDOvariables[i].pdos, mv->pdos, sizeof(TPDOvariables[i].pdos ));
      ode = od->seek(mv->index, mv->subindex);
      if( ode != NULL )
        {
          strncpy(TPDOvariables[i].variableName, ode->parameterName, 64 );        
          TPDOvariables[i].dataFormat =  canFormatToTineFormat(ode->dataTypeCode);
        }
      else 
        {
	   Lg << ERROR << "--Error: node 0x" << hex << getNodeId() <<", can not find a mapped variable (TPDO) index = 0x"  << mv->index 
           << ", subindex = 0x" << mv->subindex << dec << " in the local OD." << endl; 
          strcpy(TPDOvariables[i].variableName, "UNKNOWN VARIABLE"); 
          TPDOvariables[i].dataFormat = CF_NULL;
        }
      Lg << DBG << "TPDO var(" << i << "):" << endl;
      mv->dump();      
    }
    
  Lg << DBG << endl << "Node = 0x" << hex << nodeId << dec << ": RPDO automatic variables: " << endl;
  n = pdoClient->getRpdoVariablesCount(); 
  for( i = 0; i < n; i++ )
    { 
      mv = pdoClient->getRmappedVariable(i);


      RPDOvariables[i].index = mv->index;
      RPDOvariables[i].subindex = mv->subindex;
      RPDOvariables[i].dataValue = 0; //
      RPDOvariables[i].updateTime = 0; //
      RPDOvariables[i].dataSize = mv->dataSize;
      RPDOvariables[i].npdos = mv->npdos;
      memcpy( RPDOvariables[i].pdos, mv->pdos, sizeof(RPDOvariables[i].pdos));
      ode = od->seek(mv->index, mv->subindex);
      if( ode != NULL )
        {
          strncpy(RPDOvariables[i].variableName, ode->parameterName, 64 );        
          RPDOvariables[i].dataFormat =  canFormatToTineFormat(ode->dataTypeCode);
        }
      else 
        {
	  Lg << ERROR << "--Error: node 0x" << hex <<  getNodeId() << ",can not find a mapped variable (RPDO) index = 0x"  << mv->index 
           << ", subindex = 0x" << mv->subindex << dec << " in the local OD." << endl; 
          strcpy(RPDOvariables[i].variableName, "UNKNOWN VARIABLE"); 
          RPDOvariables[i].dataFormat = CF_NULL;
        }


      Lg << DBG << "RPDO var(" << i << "):" << endl;
      mv->dump();
    }

  n = pdoClient->getTpdoVariablesCount();
  for( i = 0; i < n; i++ )addAutomaticProperty( pdoClient->getTmappedVariable(i) );

  n = pdoClient->getRpdoVariablesCount();
  for( i = 0; i < n; i++ )addAutomaticProperty( pdoClient->getRmappedVariable(i) );   
  
}

void CanNodeImage::addAutomaticProperty( MappedVariable * mv )
{ 
  TineProperty p; 

  if( getAutomaticProperty( mv, &p ) == 0 )        
    {
      Lg << DBG << "Exporting for node 0x" << hex << getNodeId() << ": index = 0x" << mv->index << ", subindex = 0x" << mv->subindex << dec << ", name = " << p.getName() << endl;     
      addProperty( &p,  &CanNodeImage::defaultAutomaticPropertyHandler );
    }
  else Lg << ERROR << "--Error: node 0x" << hex << getNodeId() << ", getAutomaticProperty(0x" << mv->index << ",0x" << mv->subindex  << ") failed." << dec << endl;
}

int  CanNodeImage::getAutomaticProperty(MappedVariable *mv, TineProperty * p )
{
  ODentry * oe; 
  DTYPE dtype;
  
  oe = od->seek(mv->index, mv->subindex);
  if(oe == NULL)
    {
      Lg << ERROR << "--Error: node 0x" << hex << getNodeId() << ", internal OD and PDO mapping mismatch. Index = 0x" << mv->index 
           << ", subindex = 0x" << mv->subindex << dec << endl; 
      return -1;
    }

  p->setName( oe->parameterName );  //here some name modification needed for reasonable property name
  dtype.dArrayLength = 0; 
  dtype.dFormat = CF_NULL;
  p->setDataInInfo( &dtype ); 
  dtype.dArrayLength = 1;
  dtype.dFormat = canFormatToTineFormat(oe->dataTypeCode);
  p->setDataOutInfo( &dtype );
  p->setDataInInfo( &dtype );
  p->setAccess( ( oe->accessTypeCode & WRITE ) ? (CA_READ | CA_WRITE) : CA_READ );

  //  p->automaticPropertyObject = oe->getCurrentValuePtr();
  //p->mappedVariable = mv;
  p->setUserData( mv );

  sprintf( p->getDescription(), "Index=0x%X, Subindex=%d", mv->index, mv->subindex);
  return 0;
}

int CanNodeImage::canFormatToTineFormat(unsigned int canFormat )
{

  if( canFormat >= sizeof( CanToTineFormat ) ) return CF_NULL;
  return  CanToTineFormat[canFormat];
}

void CanNodeImage::setNodeId( unsigned int id )
{ 
  nodeId = id; 
  sdoClient->setNodeId(id); 
  pdoClient->setNodeId(id); 
  nodeGuardProducer->setOwnerId(id);
}

void CanNodeImage::addProperty( TineProperty * p, int (CanNodeImage::*h)( char * propertyName,  TINE_ARGS ))
{
  strcpy( propertyHandlerList[propertyCounter].name, p->getName());
  propertyHandlerList[propertyCounter].handler = h;
  TineDevice::addProperty(p->getName(), p->getDataOutInfo(), p->getDataInInfo(), p->getAccess(), p->getDescription(), p->getUserData());
}

void CanNodeImage::addProperty( char * propName, DTYPE* out, DTYPE* in, int acc, char * dsc, 
                                int (CanNodeImage::*h)( char *, TINE_ARGS ), 
                                MappedVariable *mappedVariable )
{
  strcpy( propertyHandlerList[propertyCounter].name, propName);
  propertyHandlerList[propertyCounter].handler = h;
  TineDevice::addProperty(propName, out, in, acc, dsc, mappedVariable);
} 

int CanNodeImage::propertyHandler( char *property, TINE_ARGS )
{
  for( int i = 0; i < propertyCounter; i++ )        
    {
      if( !strcmp( property, propertyHandlerList[i].name) ) return ( this->*propertyHandlerList[i].handler)( property, dout, din, access ); 
    }

  return 0;
}

int CanNodeImage::defaultAutomaticPropertyHandler( char *propertyName, TINE_ARGS )
{
  TineProperty *p;
  DTYPE * dInfo;
  if(  heartBeatConsumer->getNodeState() != OPERATIONAL ) return illegal_mode;

  if((p = getProperty(propertyName)) == NULL) return illegal_property ;
   
  if( access & CA_WRITE )
    {       
      if( din == NULL ) return illegal_data_size ; 
      dInfo = p->getDataInInfo();
      if( din->dFormat != dInfo->dFormat  ){ Lg << DBG << "expected format = " << dInfo->dFormat <<", got: " <<  din->dFormat<< endl; return illegal_format;}
      if( din->dArrayLength != dInfo->dArrayLength ) return illegal_data_size;     
      //pdoClient->sendMappedVariable( p->mappedVariable, din->data.vptr );
      pdoClient->sendMappedVariable( (MappedVariable*)(p->getUserData()), din->data.vptr );
    }
  if( access & CA_READ )
    {
      unsigned int outFormatSize =  GetFormatSize(dout->dFormat);
      if( dout == NULL ) return illegal_data_size ; 
      dInfo = p->getDataOutInfo();
      if( outFormatSize < (unsigned int) GetFormatSize(dInfo->dFormat) ) return illegal_format;
      
      putDataTimeStamp( 0, ((MappedVariable*)(p->getUserData()))->updateTime / 1000, ((MappedVariable*)(p->getUserData()))->updateTime % 1000 );
      memcpy( dout->data.bptr, ((MappedVariable*)(p->getUserData()))->dataPtr, ((MappedVariable*)(p->getUserData()))->dataSize);             
      if( dout->dArrayLength > 1 ) 
        {
          memset(  dout->data.bptr + outFormatSize, '\0', outFormatSize );                
          memcpy( dout->data.bptr + outFormatSize, &(((MappedVariable*)(p->getUserData()))->updateTime), outFormatSize < sizeof(long long) ? outFormatSize : sizeof(long long));           
          if( dout->dArrayLength > 2 ) 
            {
              memset(  dout->data.bptr +2 * outFormatSize, '\0', outFormatSize );
              *((int*)(dout->data.bptr +2 * outFormatSize)) = p->getDataOutInfo()->dFormat;
            }       
        }
    }
  return 0;
}

void CanNodeImage::activate()
{
  TimeSlice::activate();

  if( !ready )
    {
      init();
      return;  
    }

  sdoForwarder->setBlocked( flashProgrammer->isActive() );  // SDOforwarder can not request SDO channel when flash is being programmed.

  switch( heartBeatConsumer->getNodeState()  )
    {
    case INIT:
      break;
    case BOOT_UP:
      //bootup();
      break;
    case PREOPERATIONAL:
      preoperational();
      break;
    case OPERATIONAL:
      operational();
      break;
    case STOPPED:
      stopped();
      break;
    default: 
      if( !(heartBeatConsumer->getNodeState() & HEARTBEAT_TIMEOUT)) 
        Lg << ERROR << "--Error: node 0x" << hex << getNodeId() << ", internal error, unknown state = 0x" << heartBeatConsumer->getNodeState() << dec << endl;
      break;
    }

  if( needInitialization == 2 )
    {
      if( sequencer->done() )
        {                           
          sequencer->reset();
          needInitialization = 3;
        }
    }
}

void CanNodeImage::init()
{
  if( ready ) return;
  if( sequencer->available() ) configureRemoteNode();
  if( sequencer->done() )
    {
      sequencer->reset();
      heartBeatConsumer->start();
      ready = 1;
    }
}

void CanNodeImage::bootup()
{
  Lg << INFO << "NMT BOOT_UP message detected for node 0x" << hex << nodeId << dec << endl;  

  ready = 0;
  init();
  

  /* if( needInitialization == 0 )
    {
      Lg << "INIT/BOOT_UP, requesting preoperational state for node 0x" << hex << nodeId << dec << endl;
      sendNMT( ENTER_PREOPERATIONAL );  
       
      Lg << "Request to node configuration: starting sequencer." << endl;  

      pdoClient->reset();  // remove old mappings, if any
      Sequence * initSequence = new Sequence( NULL );
      initSequence->add( &CanNodeImage::seqCheckDeviceType );
      initSequence->add( &CanNodeImage::seqCheckNodeGuardNeeded );       
      initSequence->add( &CanNodeImage::seqReadTpdoMapping );       
      initSequence->add( &CanNodeImage::seqReadRpdoMapping );   
      initSequence->add( &CanNodeImage::seqExportPdoVariables );  

      sequencer->set(initSequence);
      sequencer->run();
      needInitialization = 2;
      } */
}
void CanNodeImage::preoperational()
{   
  if( needInitialization == 3 ) needInitialization = 0;
}

void CanNodeImage::operational()
{
}

void CanNodeImage::stopped()
{
}

char *  CanNodeImage::sdoBufferToString( char * data, int dataType, int dataSize )
{
    static char buffer[520];
    
    switch( dataType )
    {
	case CO_BOOL1:
	if( *(data) ) strcpy( buffer, "TRUE" );
	else strcpy(buffer, "FALSE");
	break;
      
	case CO_INT8: 
	case CO_UINT8:
	    sprintf( buffer, "%d (0x%X)", (int)(*((char*)data)), (int)(*((char*)data)));
	    break;

	case CO_INT16:  
	case CO_UINT16:
	    sprintf(buffer, "%d (0x%X)", (short)(*((short*)data )),(short)(*((short*)data )));
	    break;

	case CO_INT32:
	case CO_UINT32: 	
	    sprintf( buffer, "%d (0x%X)", (int)(*((int*)data)), (int)(*((int*)data)));
	    break;

	case CO_INT24:
	case CO_UINT24:
	    sprintf( buffer, "%d, (0x%X)", (int)((*((int*)data)) & 0x0FFFFFF ), (int)((*((int*)data)) & 0x0FFFFFF ));
	    break;  

	case CO_REAL32: 
	    sprintf( buffer, "%f", (float)(*((float*)data)));
	    break;    

	case CO_REAL64:    
	    sprintf( buffer, "%f", (double)(*((double*)data)));
	    break;  

	case CO_INT40:  
	case CO_UINT40:
	case CO_INT48: 
	case CO_UINT48: 
	case CO_INT56: 
	case CO_UINT56: 
	case CO_INT64: 
	case CO_UINT64:
	    sprintf( buffer, "%lld (0x%llX)", (long long int)(*((long long int*)data)), (long long int)(*((long long int*)data)));     
	    break; 

	case CO_VIS_STR:
	    strncpy(buffer,data, dataSize);
	    buffer[dataSize] = '\0';
	    break;

	case CO_OCTET_STR: 
	case CO_BIT_STR:
	case CO_UNICODE_STR:
	case CO_TIME_OF_DAY:
	case CO_TIME_DIFF:
	case CO_DOMAIN:     
	    sprintf( buffer, "{raw bytes} %s", data );
	    break;
        default:
            sprintf( buffer, "{unknown data type} %d (0x%X)", (int)(*((int*)data)), (int)(*((int*)data)));
            break;
    }
    return buffer;
}

void CanNodeImage::dumpSDOdataRead()
{
  Lg << DBG << "Dump of SDO data read" << endl;
  Lg << DBG << "Index = 0x" << hex << sdoClient->getIndex() << ", subindex = 0x" << sdoClient->getSubindex() << dec << endl;
  Lg << DBG << "DataType = " << sdoClient->getDataType() << ", access = " << sdoClient->getAccess() << endl;
  Lg << DBG << "Data = " << sdoBufferToString(sdoClient->getBuffer(),sdoClient->getDataType(), sdoClient->getDataSize() );  
}

int CanNodeImage::emcyHandler( int, CanMessage& m )
{
  unsigned char * bytes = m.getData();

  if( m.getDataLen() == 0 ) return nmtHandler(0,m); // some old devices send (COBID = (0x80 + $NODE_ID) and no data) as a boot-up message
        
  if( m.getDataLen() < 8 ) 
    {
      Lg << ERROR << "--Error: node = 0x" << hex << nodeId << ", wrong size of emergency message: " << dec <<  m.getDataLen() << endl;
      return -1;
    }

  int emcyCode = (((int)bytes[1]) << 8) |  bytes[0];
  int errorRegister = bytes[2];

  emcyList.add(emcyCode, errorRegister, (char*)( &bytes[3] ));
  
  Lg << INFO << "EMERGENCY MESSAGE: node = 0x" << hex << nodeId << ", Error register = 0x" << errorRegister << ", emcyCode = 0x" << emcyCode << dec << endl;
  Lg << INFO << "Manufacture specific code: " << hex ;
  for( int i = 0; i < 5; i ++ ) Lg << INFO << " 0x" << (int)bytes[3+i];
  Lg << INFO << dec << endl;

  Lg << INFO << "Error register: " << getErrorRegisterText( errorRegister ) << ", Emergency message: " << getEmcyText(emcyCode) << endl;
  return 0;
}

int CanNodeImage::sdoHandler( int, CanMessage& m )
{
  return sdoClient->handle( m );
}

int CanNodeImage::pdoHandler( int pdoNr,  CanMessage& m )
{
  return pdoClient->handle( pdoNr, m );
}

int CanNodeImage::nmtHandler(int, CanMessage& m)
{ 
  if( !ready ) return -1;   // ignore NMT messages untill init phase is finished.
  if( m.isRTR() ) return -1;  // ignore RTR frames
  if( m.getDataLen() == 0 ) // some old devices send (COBID = (0x80 + $NODE_ID) and no data) as a boot-up message
    {
       bootup();
       heartBeatConsumer->setNodeState(BOOT_UP);
      return 0;
    }

  
  // handle immediatelly
  switch (m.getData()[0] & 0x7F )  // remove 0x80 - a toggling bit for Node Guard response
    {
    case BOOT_UP: bootup();
      break;
    case PREOPERATIONAL: preoperational(); 
      break;
    case OPERATIONAL: operational(); 
      break;
    case STOPPED: stopped(); 
      break;
    case INIT: init();
      break;
    default:
      break;
    } 
  // update heart beat consumer
  heartBeatConsumer->setNodeState( m.getData()[0] );
  return 0;
}


int CanNodeImage::parse( char * fileName )
{
  int retCode = EdsParser->parse( (void*)this, fileName,  sectionParsingHandlerWrapper, objectDictionaryParsingHandlerWrapper );
  if( !retCode ) analyseOD();
  else Lg << WARNING << "--Warning: parsing of " << fileName << " failed." << endl;
  return retCode;
}

int CanNodeImage::parse( char * fileName, char *nodeName, int nodeID )
{
  int retCode = EdsParser->parse( (void*)this, fileName,  sectionParsingHandlerWrapper, objectDictionaryParsingHandlerWrapper );
  
  setName(nodeName); // overwrite node name and ID with given parameters
  setNodeId(nodeID);
  if( !retCode ) analyseOD();
  else Lg << WARNING << "-Warning: parsing of " << fileName << " failed." << endl;
  return retCode;
}

void CanNodeImage::analyseOD()
{
  ODentry * ode;
  
  //1. fix the $NODEID 
  od->goTop(); 
  ode = od->getCurrent();
  do
    {       
      if( ode->defaultValue.useNodeIDoffsetUsed() ) ode->defaultValue.v.valUINT32 +=  nodeId;
    }  while( (ode = od->getNext()) != NULL );
  // some NodeImage customization now...
  // 2. is Node Guard needed ? Check the entry: 0x100E, Node Guard Identifier
  if( od->seek( 0x1017, 0 ) == NULL ) 
    {
      if( ((ode = od->seek( 0x100E, 0 )) == NULL) )
      {
          Lg << WARNING << "-Warning: no hearbeat nor nodeguard entry found in the OD for node: " << getNodeId() << endl;
      }
      else
      //nodeGuardProducer->setId( ode->defaultValue.v.valUINT32 + ( ode->defaultValue.useNodeIDoffsetUsed() ? nodeId : 0 ));
      nodeGuardProducer->setId( ode->defaultValue.v.valUINT32 );
      nodeGuardProducer->enable();
    }
  // 3.... 

} 

void CanNodeImage::configureRemoteNode()
{

  Lg << DBG << "configuring remote node 0x" << hex << getNodeId() << dec <<endl;  
  Lg << DBG << "requesting PREOPERATIONAL state for node 0x" << hex << getNodeId() << dec <<endl;  
  
  sendNMT( ENTER_PREOPERATIONAL ); 
  heartBeatConsumer->stop();
  pdoClient->reset(); 
  
  heartBeatConsumer->setNodeState( INIT );

  Sequence * confSequence = new Sequence();

  if( ignoreBootup ) ignoreBootup = !ignoreBootupTimer->expired();

  if( ignoreBootup ) Lg << INFO << "BOOTUP IGNORED, FLASH REPROGRAMMING?" << endl;
  else
  {
      confSequence->add( &CanNodeImage::seqCheckDeviceType );
      confSequence->add( &CanNodeImage::seqSetHeartBeatTime );

      confSequence->add( &CanNodeImage::seqDisableRpdo );
      confSequence->add( &CanNodeImage::seqDisableTpdo );
      confSequence->add( &CanNodeImage::seqDisableRpdoMapping );
      confSequence->add( &CanNodeImage::seqDisableTpdoMapping );
      confSequence->add( &CanNodeImage::seqSetRpdoMapping );
      confSequence->add( &CanNodeImage::seqSetTpdoMapping );
      confSequence->add( &CanNodeImage::seqEnableRpdoMapping );
      confSequence->add( &CanNodeImage::seqEnableTpdoMapping );
      confSequence->add( &CanNodeImage::seqSetRpdoParam );
      confSequence->add( &CanNodeImage::seqSetTpdoParam );
      confSequence->add( &CanNodeImage::seqEnableRpdo );
      confSequence->add( &CanNodeImage::seqEnableTpdo );

      confSequence->add( &CanNodeImage::seqCheckNodeGuardNeeded );
      confSequence->add( &CanNodeImage::seqReadTpdoConfig );
      confSequence->add( &CanNodeImage::seqReadTpdoMapping );
      confSequence->add( &CanNodeImage::seqReadRpdoConfig );
      confSequence->add( &CanNodeImage::seqReadRpdoMapping );
      confSequence->add( &CanNodeImage::seqSetCurrentValuesForRpdoMappedVariables );
      confSequence->add( &CanNodeImage::seqExportPdoVariables );
  }
  confSequence->add( &CanNodeImage::seqReadDefaultEntries );
  confSequence->add( &CanNodeImage::seqStoreDefaultEntriesForTine);
  
  sequencer->set(confSequence);
  Lg << DBG << "Starting sequenecer for node 0x "<< hex << getNodeId() << dec << endl;
  sequencer->run(); 
}

void CanNodeImage::sectionParsingHandlerWrapper( void * object, char * sectionName, char * keyName, char *keyValue )
{
  ((CanNodeImage *)object)->sectionParsingHandler( sectionName, keyName, keyValue );
}

void CanNodeImage::objectDictionaryParsingHandlerWrapper( void * object, int index, int subindex, char * keyName, char *keyValue, int lineNumber )
{
  ((CanNodeImage *)object)->objectDictionaryParsingHandler( index, subindex, keyName, keyValue, lineNumber );
}

void  CanNodeImage::sectionParsingHandler( char * sectionName, char * keyName, char *keyValue )
{
  storeTempBufferEntry();  // if the section is changed, then store the previous object dictionary entry (if it was an OD entry)

  if( !strcmp(sectionName, "" )) return;

  if( !strcmp( sectionName, "DEVICECOMISSIONING" ) )
    {
      if( !strcmp( keyName, "NODEID" ) )
        {
          setNodeId( strtol( keyValue, NULL, 0 ) );
          return;
        }

      if( !strcmp( keyName, "NODENAME" ) )
        {
          setName( keyValue );
          return;
        }

      if( !strcmp( keyName, "BAUDRATE" ) )
        { 
          baudRate = strtol( keyValue, NULL, 0 ) * 1000;     
          return;
        }
      Lg << WARNING << "-Warning: section " << sectionName << " unknown keyName: " << keyName << endl;
      return;
    }
  // other sectionName cases come here...

  // Lg << "sectionParsingHandler called, sectionName = " << sectionName << ", keyName = " << keyName << ", keyValue = " << keyValue << endl; 
}

void CanNodeImage::storeTempBufferEntry()
{
  ODentry entry;
  ODentry *e;
  int objectTypeCode;

  if( tempBuffEntry.index < 0 ) return; // buffer does not contain any valid data
 
  entry.index = tempBuffEntry.index;
  entry.subindex = tempBuffEntry.subindex;
  if( (objectTypeCode = strtol( tempBuffEntry.objectTypeCode, NULL, 0 )) == 7 ) // VAR
    {
      if( entry.subindex < 0 )  entry.subindex = 0;  // for single variable without subindex assume subindex = 0; subindex = -1 for array/record entry section
    } 

  if( (e = od->seek( entry.index, entry.subindex )) == NULL ) e = od->add(&entry);

  strcpy( e->parameterName, tempBuffEntry.parameterName );
  e->objectTypeCode = objectTypeCode;
  e->dataTypeCode = strtol(tempBuffEntry.dataTypeCode, NULL, 0 );      
  e->subNumber = strtol( tempBuffEntry.subNumber, NULL, 0 );  
  if( ( e->accessTypeCode = od->getAccessTypeCode(tempBuffEntry.accessTypeCode)) == 0 )
    {
      e->accessTypeCode = READ;
      if( e->subNumber == 0 ) 
        {
          // log warning only for not an array-head entry
          Lg << WARNING << "-Warning: problem in EDS/DCF, section lines: " << tempBuffEntry.fileLine_start << "," << tempBuffEntry.fileLine_end << endl;
          Lg << WARNING << "-Warning: node = " <<  getNodeId() <<", index = 0x" << hex <<  e->index << ", subindex = " << dec <<  e->subindex <<" unrecognized access type:" <<(int) tempBuffEntry.accessTypeCode << ", assumed READ" << endl;          
        }
    }  

  e->objectTypeCode = strtol(tempBuffEntry.objectTypeCode, NULL, 0 );
  e->PDOmappable = strtol( tempBuffEntry.PDOmappable, NULL, 0 );
  e->storeLowLimit( tempBuffEntry.lowLimit );
  e->storeHighLimit( tempBuffEntry.highLimit ); 
  e->storeDefaultValue( tempBuffEntry.defaultValue );
  e->storeDcfParameterValue( tempBuffEntry.dcfParameterValue );
  tempBuffEntry.index = tempBuffEntry.subindex = -1;
  tempBuffEntry.fileLine_end = tempBuffEntry.fileLine_start = -1;
} 

void * CanNodeImage::getODcurrentValuePtr( int index, int subindex )
{
  ODentry * oe = od->seek( index, subindex );
  if( oe )  return oe->getCurrentValuePtr();
  return NULL;
}

void  CanNodeImage::objectDictionaryParsingHandler( int index, int subindex, char * keyName, char *keyValue, int lineNo )
{

  if( (index != tempBuffEntry.index ) ||  (subindex != tempBuffEntry.subindex) )
    {
      tempBuffEntry.fileLine_end = lineNo -2;
      storeTempBufferEntry();
      tempBuffEntry.index = index;
      tempBuffEntry.subindex = subindex;
      tempBuffEntry.accessTypeCode[0] = '\0';
      tempBuffEntry.fileLine_start = lineNo -1;
    }

  if( !strcmp( keyName, "PARAMETERNAME" ))
    {
      strcpy( tempBuffEntry.parameterName, keyValue );
      return;
    }

  if( !strcmp( keyName, "OBJECTTYPE" ))
    {
      strcpy( tempBuffEntry.objectTypeCode, keyValue );
      return;
    }  

  if( !strcmp( keyName, "DATATYPE" ))
    {
      strcpy( tempBuffEntry.dataTypeCode, keyValue );     
      return;
    }  

  if( !strcmp( keyName, "ACCESSTYPE" ))
    {
      strcpy( tempBuffEntry.accessTypeCode, keyValue ); 
      return;
    }  

  if( !strcmp( keyName, "DEFAULTVALUE" ))
    {
      strcpy( tempBuffEntry.defaultValue, keyValue );
      return;
    }  
  if( !strcmp( keyName, "PDOMAPPING" ))
    {
      strcpy(tempBuffEntry.PDOmappable, keyValue);
      return;
    }  

  if( !strcmp( keyName, "SUBNUMBER" ))
    {
      strcpy( tempBuffEntry.subNumber, keyValue );
      return;
    }  

  if( !strcmp( keyName, "LOWLIMIT" ))
    {
      strcpy( tempBuffEntry.lowLimit, keyValue );
      return;
    }  

  if( !strcmp( keyName, "HIGHLIMIT" ))
    {
      strcpy( tempBuffEntry.highLimit, keyValue );
      return;
    }  

  if( !strcmp( keyName, "OBJFLAGS" ))
    {
      Lg << WARNING << "-Warning: objectDictionaryParsingHandler() :  keyname = '" << keyName << "', Index = " << index << " is NOT supported." << endl;
      return;
    }  

  if( !strcmp( keyName, "COMPACTSUBOBJ" ))
    {
      strcpy( tempBuffEntry.objectTypeCode, keyValue );
      return;
    } 

  if( !strcmp( keyName, "PARAMETERVALUE" ))
    {
      strcpy( tempBuffEntry.dcfParameterValue, keyValue );
      return;
    } 
 

  Lg << WARNING << "-Warning: objectDictionaryParsingHandler() : unknown keyname = " << keyName << " Index = " << index << ". Ignored." << endl;

}

int CanNodeImage::sendNMT( int state )
{
  char bytes[2];
  bytes[0] = state;
  bytes[1] = nodeId;
  CanMessage m( 0, bytes, 2 );  
  return  canMessageCollector->submit(m); 
}


//--------------------------------------[EMCY]------------------------------------------------
char * CanNodeImage::getErrorRegisterText( unsigned char c )
{
  static char errorRegisterText[256];  
  if( !c ) return "no error";

  errorRegisterText[0] = '\0';
  if( c & 0x01 ) strcat( errorRegisterText, "generic," );
  if( c & 0x02 ) strcat( errorRegisterText, "current," );
  if( c & 0x04 ) strcat( errorRegisterText, "voltage," );
  if( c & 0x08 ) strcat( errorRegisterText, "temperature," );
  if( c & 0x10 ) strcat( errorRegisterText, "communication," );
  if( c & 0x20 ) strcat( errorRegisterText, "device profile specific," );
  if( c & 0x40 ) strcat( errorRegisterText, "<reserved>," );
  if( c & 0x80 ) strcat( errorRegisterText, "manufacturte specific" );

  return errorRegisterText;
}


char * CanNodeImage::getEmcyText( unsigned short c )
{ 
  switch( (c & 0xFF00) >> 8 )
    {
    case 0x00: return "Error reset, no error";
      break;
    case 0x10: return "Generic";
      break;
    case 0x20: return "Current";
      break;
    case 0x21: return "Current - device inputs";
      break;
    case 0x22: return "Current - device the device";
      break;
    case 0x23: return "Current - device outputs";
      break;
    case 0x30: return "Voltage";
      break;
    case 0x31: return "Voltage - mains voltage";
      break;
    case 0x32: return "Voltage - inside the device";
      break;
    case 0x33: return "Voltage - output";
      break;
    case 0x40: return "Temperature";
        break;
    case 0x41: return "Temperature - ambient";
      break;
    case 0x42: return "Temperature - Device";
      break;
    case 0x50: return "Device hardware";
      break;
    case 0x60: return "Device software";
      break;
    case 0x61: return "Device software - internal";
      break;
    case 0x62: return "Device software - user";
      break;
    case 0x63: return "Device software - data set";
      break;
    case 0x70: return "Additional modules";
      break;
    case 0x80: return "Monitoring";
      break;
    case 0x81: 
      switch( c & 0xFF )
        {
        case 0x10: return "Monitoring - CAN Overrun (objects lost)";
          break;
        case 0x20: return "Monitoring - CAN in error passive mode";
          break;       
        case 0x30: return "Monitoring - Node Guarding or Heart Beat Error";
          break;
        case 0x40: return "Monitoring - recovering from bus off";
          break;       
        case 0x50: return "Monitoring - COB ID";
          break;          
        default: return "Monitoring - communication";
          break;
        }
      break;
    case 0x82:
      switch( c & 0xFF )
        {
        case 0x10: return "Protocol - PDO not processed due to length error";
          break;
        case 0x20: return "Protocol - PDO length exceeded";
          break;
        default:   return "Protocol";
          break;
        }
      break;
    case 0x90: return "External";
      break;
    case 0xF0: return "Additional functions";
      break;
    case 0xFF: return "Device specific";
      break;
    default: 
      break;      
    }
  return "Unknown error";
}

//--------------------------------[Sequencer step functions]----------------------------------------

int CanNodeImage::submitSDOread( int index, int subindex, int dtype )
{
  return submitSDOread( index, subindex, buffer, dtype );
}

int CanNodeImage::submitSDOread( int index, int subindex, void *data, int dtype )
{
  int errCode;

  if( (errCode = sdoClient->read( index, subindex, data, dtype )) != NO_ERROR )
    {
      
      if( errCode == ERROR_SDO_BUSY ) 
      {
	  Lg << DBG << "Node = 0x" << hex << nodeId << ", SDOclient busy." << dec << endl;
	  return SEQUENCER_RETRY;
      }
      Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->read returned error: " << sdoClient->getErrorDescription( errCode ) <<", 0x" << errCode << dec << endl;     
      return SEQUENCER_ABORT;
    }
  return SEQUENCER_CONTINUE;
}

int CanNodeImage::submitSDOwrite( int index, int subindex, void *data, int dtype )
{
  int errCode;

  if( dtype == CO_UINT16) Lg << DBG << "Node = 0x" << hex << nodeId << ", submiting SDO write for index = 0x" << index << ", subindex 0x" << subindex << ", value = 0x" << *((unsigned short*)data) << dec << endl;
  else Lg << DBG << "Node = 0x" << hex << nodeId << ", submiting SDO write for index = 0x" << index << ", subindex 0x" << subindex << ", value = 0x" << *((unsigned long*)data) << dec << endl;
  if( (errCode = sdoClient->write( index, subindex, data, dtype )) != NO_ERROR )
    {
      if( errCode == ERROR_SDO_BUSY )
      {
	  Lg << DBG << "Node = 0x" << hex << nodeId << ", SDOclient busy." << dec << endl;
	  return SEQUENCER_RETRY;
      }
      Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode ) <<", 0x" << errCode << dec << endl;
      return SEQUENCER_ABORT;
    }
  return SEQUENCER_CONTINUE;
}

int CanNodeImage::seqSetHeartBeatTime(int *callNo, void *a)
{
  __PRINT_FUNCTION_NAME__
  ODentry * ode;
  if( (*callNo)++ == 0 )
    {
      ode = od->seek( 0x1017, 0 );
      if( ode == NULL ) 
        {
          Lg << WARNING << "-Warning: no Producer Heartbeat Time entry in the object dictionary for node: 0x"<< hex << getNodeId() << dec << endl;
          return SEQUENCER_NO_ERROR;
        }
      *((ODentry**)a) = ode;
      short hbt = (ode->getDefaultValuePtr())->valINT16;
      return submitSDOwrite( 0x1017, 0, &hbt,  CO_UINT16 );
    } 
  if( sdoClient->done() )
    {
      // copy always default to current heart beat value, even the node is not responding if( sdoClient->getError() == NO_ERROR )         
      (*((ODentry**)a))->copyDefaultToCurrentValue();
      heartBeatConsumer->setPeriod( 2 * ((*((ODentry**)a))->getDefaultValuePtr())->valINT16 );        
      return checkSDOwrite();      
    }
  
  return SEQUENCER_CONTINUE;
}

int CanNodeImage::checkSDOwrite()
{
  if( !sdoClient->done() ) return SEQUENCER_CONTINUE;
  
  int errCode;
  if( (errCode = sdoClient->getError()) != NO_ERROR) 
    {
      Lg << ERROR << "--Error: node 0x" << hex << getNodeId() << ", SDOclient->write( 0x" <<  sdoClient->getIndex() <<", " 
         << sdoClient->getSubindex() << ")  ACK message returned error: " << sdoClient->getErrorDescription( errCode ) 
         <<", 0x" << errCode << dec << endl; 
      if( errCode == ERROR_ABORT ) Lg << ERROR << sdoClient->getAbortDescription() << endl;
    }
  sdoClient->reset();
  return SEQUENCER_NO_ERROR;  
}

int CanNodeImage::seqDisableRpdo( int *callNo, void *a )
{ 
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;
  __PRINT_FUNCTION_NAME__

  if( (*callNo)++ == 0)
    {
      p->skipSubindex = -1;
      p->startSubindex = 0x01;
      p->endIndex = 0x14FF;
      p->endSubindex = 0x01;
      p->value =  getNodeId() | 0x80000000;
      p->type =  CO_UINT32;
      if( (p->ode = findNextWritableIndex(0x1400 -1, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;  // no writable entry in range 0x1400 .. 0x14FF
      p->startIndex = p->ode->index;      
      return submitSDOwrite( p->startIndex, p->startSubindex, &p->value, p->type );
    }
  return continueDisablePdo(p);
}

int CanNodeImage::continueDisablePdo( OD_WRITE_SEQUENCE_PAR *p )
{
  if( checkSDOwrite() == SEQUENCER_CONTINUE ) return SEQUENCER_CONTINUE;
  if( p->startIndex == p->endIndex ) return SEQUENCER_NO_ERROR;  
  if( (p->ode = findNextWritableIndex(p->startIndex, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;  // no writable entry in range p->startIndex, p->endIndex
  p->startIndex = p->ode->index;
  return submitSDOwrite( p->startIndex, p->startSubindex, &p->value, p->type );
}

int CanNodeImage::seqDisableTpdo( int *callNo, void *a )
{ 
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

 __PRINT_FUNCTION_NAME__

  if( (*callNo)++ == 0)
    {
      p->skipSubindex = -1;      
      p->startSubindex = 0x01;
      p->endIndex = 0x18FF;
      p->endSubindex = 0x01;
      p->value =  getNodeId() | 0x80000000;
      p->type =  CO_UINT32;
      if( (p->ode = findNextWritableIndex(0x1800 -1, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;  // no writable entry in range 0x1800 .. 0x18FF
      p->startIndex = p->ode->index;
      return submitSDOwrite( p->startIndex, p->startSubindex, &p->value, p->type );
    }
  return continueDisablePdo(p); 
}

int CanNodeImage::seqDisableRpdoMapping( int *callNo, void *a )
{  
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

 __PRINT_FUNCTION_NAME__

  if( (*callNo)++ == 0)
    { 
      p->skipSubindex = -1;      
      p->startSubindex = 0x00;
      p->endIndex = 0x16FF;
      p->endSubindex = 0x00;
      p->value = 0;
      p->type =  CO_UINT8;
      if( (p->ode = findNextWritableIndex(0x1600 -1, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;  // no writable entry in range 0x1600 .. 0x16FF
      p->startIndex = p->ode->index;
      return submitSDOwrite( p->startIndex, p->startSubindex, &p->value, p->type );
    }
  return continueDisablePdo(p);
}
int CanNodeImage::seqDisableTpdoMapping( int *callNo, void *a )
{  
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

 __PRINT_FUNCTION_NAME__

  if( (*callNo)++ == 0)
    { 
      p->skipSubindex = -1;      
      p->startSubindex = 0x00;
      p->endIndex = 0x1AFF;
      p->endSubindex = 0x00;
      p->value = 0;
      p->type =  CO_UINT8;
      if( (p->ode = findNextWritableIndex(0x1A00 -1, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;  // no writable entry in range 0x1A00 .. 0x1AFF
      p->startIndex = p->ode->index;
      return submitSDOwrite( p->startIndex, p->startSubindex, &p->value, p->type );
    }
  return continueDisablePdo(p);
}

int CanNodeImage::seqEnableRpdoMapping( int *callno, void *a )
{  
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

 __PRINT_FUNCTION_NAME__

  if( (*callno) == 0)
    {
      p->skipSubindex = -1;      
      p->startSubindex = 0x00;
      p->endIndex = 0x16FF;
      p->endSubindex = 0x00;
      p->type =  CO_UINT8;
      if( (p->ode = findNextWritableIndex(0x1600 -1, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;  // no writable entry in range 0x1600 .. 0x16FF
      p->startIndex = p->ode->index;
    }
  return continueEnablePdoMapping(callno, p);
}

int CanNodeImage::seqEnableTpdoMapping( int *callno, void *a )
{  
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

  if( (*callno) == 0)
    {
      p->skipSubindex = -1;
      p->startIndex = 0x1A00;
      p->startSubindex = 0x00;
      p->endIndex = 0x1AFF;
      p->endSubindex = 0x00;
      p->type =  CO_UINT8;      
    }
  return continueEnablePdoMapping( callno, p);
}

int CanNodeImage::continueEnablePdoMapping(int *callno, OD_WRITE_SEQUENCE_PAR * p )
{ 

  if( (*callno)++ == 0 )
    {      
      if( (p->ode = findNextWritableIndex( p->startIndex -1, p->endIndex, p->startSubindex )) == NULL ) 
        {
          Lg << WARNING << "-Warning: no configurable RPDO mapping in DCF file?" << endl;
          return SEQUENCER_NO_ERROR;
        }
      p->value = *((long*)(p->ode->getDefaultValuePtr()));      
      p->startIndex = p->ode->index;
      return submitSDOwrite( p->startIndex, p->startSubindex, &p->value, p->type );
    }
  if( checkSDOwrite() == SEQUENCER_CONTINUE ) return SEQUENCER_CONTINUE; 
  
  if( (p->ode = findNextWritableIndex( p->startIndex, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;
  p->startIndex = p->ode->index;
  p->value =  *((long*)(p->ode->getDefaultValuePtr()));
  if( p->ode->subindex < 0 ) 
    {
      Lg << ERROR << "--Error: node 0x" << hex << nodeId << dec << ", negative subindex: " << (char*) __FUNCTION__ << ", subindex = " << p->ode->subindex << endl;
    }
  return submitSDOwrite( p->ode->index, p->ode->subindex, p->ode->getDefaultValuePtr(), p->ode->dataTypeCode );

}

int CanNodeImage::seqSetRpdoMapping( int *callNo, void *a )
{
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

__PRINT_FUNCTION_NAME__

  if( *callNo == 0) 
    {                                                            
      p->startIndex = 0x1600;  
      p->skipSubindex = 0x00;
      p->startSubindex = 0x01;
      p->endIndex = 0x16FF;
      p->endSubindex = 0xFF;          
    }

  return sendPdoConfig(callNo, p);
}

int CanNodeImage::seqSetTpdoMapping( int *callNo, void *a )
{
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

__PRINT_FUNCTION_NAME__

  if( *callNo == 0) 
    {                                                            
      p->startIndex = 0x1A00;  
      p->skipSubindex = 0x00;
      p->startSubindex = 0x01;
      p->endIndex = 0x1AFF;
      p->endSubindex = 0xFF;          
    }

  return sendPdoConfig(callNo, p);
}

int CanNodeImage::seqSetRpdoParam( int *callNo, void *a )
{
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

__PRINT_FUNCTION_NAME__

  if( *callNo == 0) 
    {                                                            
      p->startIndex = 0x1400; 
      p->skipSubindex = 0x01;
      p->startSubindex = 0x02;
      p->endIndex = 0x14FF;
      p->endSubindex = 0xFF;          
    }

  return sendPdoConfig(callNo, p);
}

int CanNodeImage::seqSetTpdoParam( int *callNo, void *a )
{
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

__PRINT_FUNCTION_NAME__

  if( *callNo == 0) 
    {                                                            
      p->startIndex = 0x1800;
      p->skipSubindex = 0x01;
      p->startSubindex = 0x02;
      p->endIndex = 0x18FF;
      p->endSubindex = 0xFF;          
    }

  return sendPdoConfig(callNo, p);
}

int CanNodeImage::seqEnableRpdo( int *callNo, void *a )
{  
  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

__PRINT_FUNCTION_NAME__

  if( (*callNo) == 0)
    { 
      p->skipSubindex = -1;
      p->startIndex = 0x1400;
      p->startSubindex = 0x01;
      p->endIndex = 0x14FF;
      p->endSubindex = 0x01;
      p->type =  CO_UINT32;
    }
  return continueEnablePdo(callNo, p);
}

int CanNodeImage::seqEnableTpdo( int *callNo, void *a )
{  

  OD_WRITE_SEQUENCE_PAR *p = (OD_WRITE_SEQUENCE_PAR*)a;

__PRINT_FUNCTION_NAME__

  if( (*callNo) == 0)
    {
      p->skipSubindex = -1;
      p->startIndex = 0x1800;
      p->startSubindex = 0x01;
      p->endIndex = 0x18FF;
      p->endSubindex = 0x01;
      p->type =  CO_UINT32;
    }
  return continueEnablePdo(callNo, p);
}

int CanNodeImage::continueEnablePdo(int *callNo, OD_WRITE_SEQUENCE_PAR * p )
{ 
__PRINT_FUNCTION_NAME__
  
  if( (*callNo)++ == 0 )
    {      
      if( (p->ode = findNextWritableIndex( p->startIndex-1, p->endIndex, p->startSubindex )) == NULL ) 
        {
          Lg << WARNING << "-Warning: no configurable PDO mapping in DCF file?" << endl;
          return SEQUENCER_NO_ERROR;
        }
      p->value = *((long*)(p->ode->getDefaultValuePtr()));            
      return submitSDOwrite( p->ode->index, p->startSubindex, &p->value, p->type );
    }
  if( checkSDOwrite() == SEQUENCER_CONTINUE ) return SEQUENCER_CONTINUE;

  if( (p->ode = findNextWritableIndex( p->ode->index, p->endIndex, p->startSubindex )) == NULL ) return SEQUENCER_NO_ERROR;    
  return submitSDOwrite( p->ode->index, p->ode->subindex, p->ode->getDefaultValuePtr(), p->ode->dataTypeCode );
}


int CanNodeImage::sendPdoConfig( int *callNo, OD_WRITE_SEQUENCE_PAR *p )
{ 
__PRINT_FUNCTION_NAME__

  if ( (*callNo)++ == 0 )
    {
      p->startSubindex--;   
      if( findNextWritableIndexSubindex(p) == NULL ) return SEQUENCER_NO_ERROR;
      return submitSDOwrite( p->ode->index, p->ode->subindex, p->ode->getDefaultValuePtr(), p->ode->dataTypeCode );
      /*
       while( p->ode == NULL || !(p->ode->accessTypeCode & WRITE) ) 
        {         
          if( p->ode == NULL )
            {               
              // entry does not exist            
              if( p->startIndex == p->endIndex ) 
                {
                  Lg << WARNING << "-Warning: no writeable PDO parameters in OD for node 0x" << hex << getNodeId() << ", endIndex = 0x" << p->endIndex << dec << endl;
                  return SEQUENCER_NO_ERROR;
                }             
              p->ode = od->seek( ++(p->startIndex), p->startSubindex = p->skipSubindex +1 );
            }
          else
            {
              // entry read only
              if( ++(p->startSubindex) == p->skipSubindex ) ++(p->startSubindex); 
              p->ode = od->seek( p->startIndex, p->startSubindex );
            }                  
        } 
      return submitSDOwrite( p->ode->index, p->ode->subindex, p->ode->getDefaultValuePtr(), p->ode->dataTypeCode );
       */
    }
 
  if( checkSDOwrite() == SEQUENCER_CONTINUE ) return  SEQUENCER_CONTINUE;
  if( findNextWritableIndexSubindex(p) == NULL ) return SEQUENCER_NO_ERROR;
 /*
  if( p->ode->subindex <= p->skipSubindex ) p->ode->subindex =  p->skipSubindex +1;
  od->seek(p->ode->index, p->ode->subindex);  // restore position in the object dictionary (somebody else can make a seek in the meantime)
  do
  {
    if( (p->ode = od->getNext() ) == NULL ) return SEQUENCER_NO_ERROR;  // no next entry in OD 
    if( p->ode->index > p->endIndex ) return SEQUENCER_NO_ERROR;
    if( p->ode->index == p->endIndex && p->ode->subindex > p->endSubindex ) return SEQUENCER_NO_ERROR;
  }
  while( !( p->ode->accessTypeCode & WRITE ) || (p->ode->subindex <= p->skipSubindex)); 
  */
  return submitSDOwrite( p->ode->index, p->ode->subindex, p->ode->getDefaultValuePtr(), p->ode->dataTypeCode );
}



int CanNodeImage::seqCheckDeviceType(int *callNo, void * a)
{

__PRINT_FUNCTION_NAME__

  if( (*callNo)++ == 0 ) return submitSDOread( 0x1000, 0, buffer, CO_UINT32 ); 
  if( sdoClient->done() )
    {
      int errCode;
      if( (errCode = sdoClient->getError()) == NO_ERROR)
        {	  
          unsigned int devType = *((unsigned int*)sdoClient->getBuffer());
	  sdoClient->reset();
          ODentry *ode = od->seek( 0x1000, 0);
          if( ode == NULL ) 
            {
              Lg << ERROR << "--Error: node = 0x" << hex << getNodeId() << dec << " object dictionary without mandatory entry 0x1000" << endl;
              return SEQUENCER_ABORT;
            }
          ode->currentValue.v.valUINT32 = devType;
          if( ode->defaultValue.v.valUINT32 != devType )
            {
              Lg << ERROR << "--Error: wrong device type for node 0x" << hex << nodeId << ", expected = 0x" << ode->defaultValue.v.valUINT32 <<", got = 0x" << devType  << dec << endl;
              setInitializationError( INIT_WRONG_DEVICE_TYPE );
              return SEQUENCER_NO_ERROR; //SEQUENCER_ABORT;
            }
          Lg << DBG << "device type OK for node 0x" << hex << nodeId << dec << endl;
        }
      else
      {
	  Lg << ERROR << "--Error, seqCheckDeviceType(), nodeId = 0x" << hex << getNodeId() << dec << ": " << sdoClient->getErrorDescription() << " (0x" << sdoClient->getError() << ")"<< endl;      
	  sdoClient->reset();      
      }
      return (errCode == NO_ERROR) ? SEQUENCER_NO_ERROR : SEQUENCER_ABORT;
    }      
  return SEQUENCER_CONTINUE;
}

int CanNodeImage::seqReadDefaultEntries(int *callNo, void * a)
{
  // reading of some entries, describing the module, like device type, hardware and software version, etc.
  DEFAULT_ENTRIES_READING * r = (DEFAULT_ENTRIES_READING *)a;
  ODentry *ode;
  if( *callNo == 0 ) 
    {
      if( (r->nItems = sizeof(DefaultUpdateEntries) / sizeof(DefaultUpdateEntries[0])) == 0 ) return SEQUENCER_NO_ERROR;  
      r->currentItem = 0;      
      ODentry * ode;
      while( (ode = od->seek( DefaultUpdateEntries[r->currentItem] )) == NULL ) if( ++(r->currentItem) == r->nItems ) return SEQUENCER_NO_ERROR;
      *callNo = 1;
      return submitSDOread( DefaultUpdateEntries[r->currentItem].index , DefaultUpdateEntries[r->currentItem].subindex, buffer, ode->dataTypeCode ); 
    }

  if( sdoClient->done() )
    {
      int errCode;
      if( (errCode = sdoClient->getError()) == NO_ERROR)
        {          
          ode = od->seek( DefaultUpdateEntries[r->currentItem] );
          if( ode == NULL ) 
            {
              Lg << ERROR << "--Error: node 0x" << hex << nodeId << dec << ", seqReadDefaultEntries internal logic." << endl; 
              sdoClient->reset(); 
              return SEQUENCER_ABORT;
            }
          ode->assign( &(ode->currentValue), sdoClient->getBuffer(), sdoClient->getDataSize() );         
        }
      else 
      {
          Lg << ERROR << "--Error: seqReadDefaultEntries, nodeId = 0x" << hex << getNodeId() << dec      
              << ": " << sdoClient->getErrorDescription() << " (" << sdoClient->getError() << ")"<< endl;
          if( errCode == ERROR_ABORT ) Lg << ERROR << sdoClient->getAbortDescription() << endl;
      }
      sdoClient->reset(); 
     
      do if( ++(r->currentItem) == r->nItems ) return SEQUENCER_NO_ERROR;
      while( (ode = od->seek( DefaultUpdateEntries[r->currentItem] )) == NULL ); 
      return submitSDOread( DefaultUpdateEntries[r->currentItem].index , DefaultUpdateEntries[r->currentItem].subindex, buffer, ode->dataTypeCode );  
    }      
  return SEQUENCER_CONTINUE;
}


#define CHECK_HEARTBEAT_ENTRY           0
#define CHECK_HEARTBEAT_ENTRY_RESPONSE  1
#define CHECK_NODEGUARD_ENTRY           2
#define CHECK_NODEGUARD_ENTRY_RESPONSE  3
#define WRITE_HEARTBEAT_ENTRY_RESPONSE  4
#define WRITE_NODEGUARD_ENTRY_RESPONSE  5

int CanNodeImage::seqCheckNodeGuardNeeded(int *callNo, void * a)
{

  __PRINT_FUNCTION_NAME__

  if(*callNo == CHECK_HEARTBEAT_ENTRY) 
    {
      *callNo = CHECK_HEARTBEAT_ENTRY_RESPONSE;
      return submitSDOread(0x1017, 0, buffer, CO_UINT16); 
    }
  if( !sdoClient->done() ) return SEQUENCER_CONTINUE;

  ODentry *ode;
  int errCode;
  switch(*callNo)
    {             
      case CHECK_HEARTBEAT_ENTRY_RESPONSE:      
          if( (errCode=sdoClient->getError()) == NO_ERROR)
          {        
             ode = od->seek(0x1017, 0);
             if( ode == NULL ) 
             {
                Lg << ERROR << "--Error: DCF mismatch for node: 0x" << hex << nodeId << dec 
                   << " : no entry for the Heartbeat, but it exists in device's Object Dictionary." << endl;
                sdoClient->reset();              
                setInitializationError( INIT_OD_NODE_MISMATCH );
                return SEQUENCER_NO_ERROR;  // error, but let the sequencer continue with the next sequenece steps
             }  
             unsigned int hb = *((unsigned int*)sdoClient->getBuffer());
             sdoClient->reset();
             if( hb != ode->defaultValue.v.valUINT16 )
             {
                *callNo = WRITE_HEARTBEAT_ENTRY_RESPONSE;
                return submitSDOwrite(0x1017, 0, &hb, CO_UINT16);  
             }
             else
             {
                 // The heartbeat entry exists and has the same value as in the DCF file.
                 // No need to start node guarding, nor anything else. 
                 return SEQUENCER_NO_ERROR;
             }
          }
           
	 sdoClient->reset();
         *callNo = CHECK_NODEGUARD_ENTRY_RESPONSE;
         return submitSDOread(0x101E, 0, buffer, CO_UINT32); 
         
      case WRITE_HEARTBEAT_ENTRY_RESPONSE: 
         if( (errCode = sdoClient->getError()) != NO_ERROR )
         {
             Lg << ERROR << "--Error: can not write the heartbeat period value for node 0x" << hex << nodeId << dec << endl;
             sdoClient->reset();
             // try to activate the node guard...
             *callNo = CHECK_NODEGUARD_ENTRY_RESPONSE;
            return submitSDOread(0x101E, 0, buffer, CO_UINT32); 
         }
         
         sdoClient->reset();
         // write ok, copy the default value to the current value  
         ode = od->seek(0x1017, 0);          
         ode->currentValue.v.valUINT16 = ode->defaultValue.v.valUINT16;      
         Lg << DBG << "Heart beat time set to: " << ode->defaultValue.v.valUINT16 << " for node 0x" << hex << nodeId << dec << endl; 
         return SEQUENCER_NO_ERROR;
         
      case CHECK_NODEGUARD_ENTRY_RESPONSE:      
      if( (errCode=sdoClient->getError()) == NO_ERROR)
        {        
         ode = od->seek(0x100E, 0);
          if( ode == NULL ) 
            {
              Lg << ERROR << "--Error: DCF mismatch for node: 0x" << hex << nodeId << dec 
                 << " : no entry for Node Guard COBID, but exists in device's Object Dictionary." << endl;
              sdoClient->reset();              
              setInitializationError( INIT_OD_NODE_MISMATCH );
              return SEQUENCER_NO_ERROR;
            }  
          unsigned int nodeGuardId = *((unsigned int*)sdoClient->getBuffer());
          sdoClient->reset();
          nodeGuardProducer->setId( nodeGuardId );     // set id that is in node's OD
          nodeGuardProducer->enable();
          ode->currentValue.v.valUINT32 = nodeGuardId;
          unsigned int id = ode->defaultValue.v.valUINT32;
          //if( ode->defaultValue.useNodeIDoffsetUsed() ) id += nodeId;          
          if( id != nodeGuardId )
            {                             
              if( sdoClient->write(0x100E, 0, &id, CO_UINT32) != NO_ERROR ) 
                {
                  sdoClient->reset();   
                  Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqCheckNodeGuardNeeded(): can not submit SDOwrite(0x100E,0)"<< dec << endl;
                  setInitializationError( INIT_SDO_SUBMIT_ERROR );
                  return SEQUENCER_ABORT;
                }                    
              *callNo = WRITE_NODEGUARD_ENTRY_RESPONSE;
              return SEQUENCER_CONTINUE;                   
            } 
          else return SEQUENCER_NO_ERROR;
        }
      else 
        {              
          if(errCode == ERROR_ABORT)
            {     
              sdoClient->reset();
              ODentry *ode = od->seek(0x100E, 0);
              if( ode == NULL )
                {
                  // consistent: no entry in node's OD, no entry in local OD                    
                  return SEQUENCER_NO_ERROR;
                }
              else
                {
                  Lg << ERROR << "--Error: DCF mismatch for node: 0x" << hex << nodeId << dec 
                     << ",entry exists for Node Guard COBID, but not in DCF." << endl;
                  setInitializationError( INIT_OD_NODE_MISMATCH );
                  return SEQUENCER_NO_ERROR;
                }  
            }
          Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqCheckNodeGuardNeeded(): " << sdoClient->getErrorDescription() << " (0x" << sdoClient->getError() << ")"<< dec << endl;
          sdoClient->reset();  
	   setInitializationError( INIT_OD_NODE_MISMATCH );
          return SEQUENCER_ABORT;        
        }
      break;
    case WRITE_NODEGUARD_ENTRY_RESPONSE:
      if( (errCode = sdoClient->getError()) != NO_ERROR)
        {        
          Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDO write error: " << sdoClient->getErrorDescription() << " (0x" << sdoClient->getError() << ")"<< dec << endl;
          if( errCode == ERROR_ABORT ) Lg << ERROR << sdoClient->getAbortDescription() << endl;          
          sdoClient->reset(); 
          setInitializationError( INIT_SDO_RW_ERROR ); 
          return SEQUENCER_ABORT;
        }
      sdoClient->reset(); 
      // write ok, copy the default value to the current value  
      ode = od->seek(0x100E, 0);    
      //ode->currentValue.v.valUINT32 = ode->defaultValue.v.valUINT32 + ( ode->defaultValue.useNodeIDoffsetUsed() ? nodeId : 0 );
      ode->currentValue.v.valUINT32 = ode->defaultValue.v.valUINT32;
      nodeGuardProducer->setId( ode->currentValue.v.valUINT32 );                   
      return SEQUENCER_NO_ERROR;
      break;
    default:
      Lg << ERROR << "--Error: nodeId 0x" << hex << nodeId << ", internal error, bad case = " << dec << *callNo << endl;
      setInitializationError( INIT_INTERNAL_ERROR );
      sdoClient->reset();
      return SEQUENCER_ABORT;
      break;
    }
  return SEQUENCER_CONTINUE;
}

typedef struct
{
  int index;
  int subindex;
  int size; 
} PDO_MAP_READOUT;

const int READ_RECORD_SIZE = 1;
const int READ_RECORD_ITEM = 2;
int CanNodeImage::seqReadTpdoMapping(int *callNo, void * a)
{
  int errCode;
  PDO_MAP_READOUT *p = (PDO_MAP_READOUT *)a;
 
  if(*callNo == 0) 
    {                                                            
      p->index = 0x1A00;
      p->subindex = 0;
      p->size = 0; 
      *callNo = READ_RECORD_SIZE;
      return submitSDOread(0x1A00, 0, buffer, CO_UINT8); 
    }
  if( !sdoClient->done() ) return SEQUENCER_CONTINUE;

  switch(*callNo)
    { 
    case READ_RECORD_SIZE:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        {  
          p->size = *((unsigned char*)sdoClient->getBuffer());           
          sdoClient->reset();                    
          if( p->size )
            {              
              *callNo = READ_RECORD_ITEM;
              return submitSDOread(p->index, ++p->subindex, buffer, CO_UINT32);                                    
            }
          else
            {
              ++p->index; 
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
            }                   
        }
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping: " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;
            return SEQUENCER_ABORT;
          }
          // no entry for the mapping. Try the next one:
          if( ++p->index <= 0x1A0F )    // try to search only first 16 entries
          {
            p->subindex = 0;
            *callNo = READ_RECORD_SIZE;
            return submitSDOread(p->index, 0, buffer, CO_UINT8);             
          }
          else return SEQUENCER_NO_ERROR;
        }      
      break;      
    case READ_RECORD_ITEM:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        {                       
          pdoClient->addTpdoMapping( p->index - 0x1A00, p->subindex-1, *((unsigned int*)sdoClient->getBuffer()));
          sdoClient->reset(); 
          if( p->subindex == p->size )
            {
              *callNo = READ_RECORD_SIZE;
              ++p->index;
              p->subindex = 0; 
              return submitSDOread(p->index,p->subindex, buffer, CO_UINT8);
            }
          *callNo = READ_RECORD_ITEM; 
          ++p->subindex;          
          return submitSDOread(p->index,p->subindex, buffer, CO_UINT32);
        }       
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {
              Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping(): " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;          
              return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x1A0F )    // try to search only first 16 entries
          {
            p->subindex = 0;
            *callNo = READ_RECORD_SIZE;
            return submitSDOread(p->index, 0, buffer, CO_UINT8);  
          }
          else return SEQUENCER_NO_ERROR;
        }
      break;
    }    
  return SEQUENCER_NO_ERROR;
}

int CanNodeImage::seqReadRpdoConfig(int *callNo, void * a)
{
  int errCode;
  PDO_MAP_READOUT *p = (PDO_MAP_READOUT *)a;
 
  if(*callNo == 0) 
    {                                                            
      p->index = 0x1400;
      p->subindex = 0;
      p->size = 0; 
      *callNo = READ_RECORD_SIZE;
      return submitSDOread(0x1400, 0, buffer, CO_UINT8); 
    }
  if( !sdoClient->done() ) return SEQUENCER_CONTINUE;

  switch(*callNo)
    { 
    case READ_RECORD_SIZE:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        {  
          p->size = *((unsigned char*)sdoClient->getBuffer());  
          sdoClient->reset();                    
          if( p->size )
            {              
              *callNo = READ_RECORD_ITEM;
              return submitSDOread(p->index, p->subindex = 1, buffer, CO_UINT32);                                    
            }
          else
            {
              ++p->index; 
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
            }                   
        }
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping: " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;
            return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x140F )
          {
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
          }
          return SEQUENCER_NO_ERROR; // no OD entry for the next mapping. Readout of mapping done. 
        }      
      break;      
    case READ_RECORD_ITEM:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        { 
          pdoClient->setRpdoCobId( p->index - 0x1400, *((unsigned int*)sdoClient->getBuffer())); 
          sdoClient->reset();          
          *callNo = READ_RECORD_SIZE;
          ++p->index;
          p->subindex = 0; 
          return submitSDOread(p->index,p->subindex, buffer, CO_UINT8);
        }                       
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {    
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping(): " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;
            return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x140F )
          {
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
          }
          return SEQUENCER_NO_ERROR; // no OD entry for the next mapping. Readout of mapping done. 
        }
      break;
    }    
  return SEQUENCER_NO_ERROR;
}

int CanNodeImage::seqReadRpdoMapping(int *callNo, void * a)
{
  int errCode;
  PDO_MAP_READOUT *p = (PDO_MAP_READOUT *)a;
 
  if(*callNo == 0) 
    {                                                            
      p->index = 0x1600;
      p->subindex = 0;
      p->size = 0; 
      *callNo = READ_RECORD_SIZE;
      return submitSDOread(0x1600, 0, buffer, CO_UINT8); 
    }
  if( !sdoClient->done() ) return SEQUENCER_CONTINUE;

  switch(*callNo)
    { 
    case READ_RECORD_SIZE:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        {  
          p->size = *((unsigned char*)sdoClient->getBuffer());  
          sdoClient->reset();                    
          if( p->size )
            {              
              *callNo = READ_RECORD_ITEM;
              return submitSDOread(p->index, ++p->subindex, buffer, CO_UINT32);                                    
            }
          else
            {
              ++p->index; 
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
            }                   
        }
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping: " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;
            return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x160F )
          {
            p->subindex = 0;
            p->size = 0; 
            *callNo = READ_RECORD_SIZE;
            return submitSDOread(p->index, 0, buffer, CO_UINT8); 
          }
          return SEQUENCER_NO_ERROR;
        }      
      break;      
    case READ_RECORD_ITEM:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        {                       
          pdoClient->addRpdoMapping( p->index - 0x1600, p->subindex-1, *((unsigned int*)sdoClient->getBuffer()));
          sdoClient->reset(); 
          if( p->subindex == p->size )
            {
              *callNo = READ_RECORD_SIZE;
              ++p->index;
              p->subindex = 0; 
              return submitSDOread(p->index,p->subindex, buffer, CO_UINT8);
            }
          *callNo = READ_RECORD_ITEM; 
          ++p->subindex;          
          return submitSDOread(p->index,p->subindex, buffer, CO_UINT32);
        }       
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping(): " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;
            return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x160F )
          {
            p->subindex = 0;
            p->size = 0; 
            *callNo = READ_RECORD_SIZE;
            return submitSDOread(p->index, 0, buffer, CO_UINT8); 
          }
          return SEQUENCER_NO_ERROR;
        }
      break;
    }    
  return SEQUENCER_NO_ERROR;
}


int CanNodeImage::seqReadTpdoConfig(int *callNo, void * a)
{
  int errCode;
  PDO_MAP_READOUT *p = (PDO_MAP_READOUT *)a;
 
  if(*callNo == 0) 
    {                                                            
      p->index = 0x1800;
      p->subindex = 0;
      p->size = 0; 
      *callNo = READ_RECORD_SIZE;
      return submitSDOread(0x1800, 0, buffer, CO_UINT8); 
    }
  if( !sdoClient->done() ) return SEQUENCER_CONTINUE;

  switch(*callNo)
    { 
    case READ_RECORD_SIZE:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        {  
          p->size = *((unsigned char*)sdoClient->getBuffer());  
          sdoClient->reset();                    
          if( p->size )
            {              
              *callNo = READ_RECORD_ITEM;
              return submitSDOread(p->index, p->subindex = 1, buffer, CO_UINT32);                                    
            }
          else
            {
              ++p->index; 
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
            }                   
        }
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadPdoMapping: " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")"<< dec << endl;
            return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x1800 +  MAX_PDO + 8)   //  assume PDOs are defined in sapce 1800 ... 1800 + MAX_PDO + 8
          {
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
          }
          return SEQUENCER_NO_ERROR; // no OD entry for the next mapping. Readout of mapping done. 
        }      
      break;      
    case READ_RECORD_ITEM:
      if((errCode=sdoClient->getError()) == NO_ERROR)
        { 
          pdoClient->setTpdoCobId( p->index - 0x1800, *((unsigned int*)sdoClient->getBuffer())); 
          sdoClient->reset();          
          *callNo = READ_RECORD_SIZE;
          ++p->index;
          p->subindex = 0; 
          return submitSDOread(p->index,p->subindex, buffer, CO_UINT8);
        }                       
      else 
        {   
          sdoClient->reset();              
          if(errCode != ERROR_ABORT) 
          {    
            Lg << ERROR << "--Error, node 0x" << hex << nodeId << ", seqReadTPdoConfig(): " << sdoClient->getErrorDescription(errCode) << " (0x" << errCode << ")" << dec << endl;
            return SEQUENCER_ABORT;
          }
          if( ++p->index <= 0x1800 + MAX_PDO + 8 )
          {
              p->subindex = 0;
              *callNo = READ_RECORD_SIZE;
              return submitSDOread(p->index, 0, buffer, CO_UINT8);
          }
          return SEQUENCER_NO_ERROR; // no OD entry for the next mapping. Readout of mapping done. 
        }
      break;
    }    
  return SEQUENCER_NO_ERROR;
}


int CanNodeImage::seqStoreDefaultEntriesForTine( int *callNo, void *a )
{
  ODvariable * var;
  ODentry * ode = od->seek( 0x1000, 0 );
  if( ode == NULL ) 
    {
      Lg << ERROR << "--Error: node 0x" << hex << nodeId << " without mandatory Device Type field" << dec << endl;
      return SEQUENCER_NO_ERROR;
    }
  var = &(ode->currentValue);
  if( var->isSet() )  canNodeIdInfo.deviceType =   var->v.valUINT32;
  ode = od->seek( 0x1018, 1 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  canNodeIdInfo.vendorId = var->v.valUINT32;
    }
  ode = od->seek( 0x1018, 2 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  canNodeIdInfo.productCode = var->v.valUINT32;
    }
  ode = od->seek( 0x1018, 3 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  canNodeIdInfo.revisionNumber = var->v.valUINT32;
    }
  ode = od->seek( 0x1018, 4 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  canNodeIdInfo.serialNumber = var->v.valUINT32;
    }
  ode = od->seek( 0x1008, 0 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  strncpy( canNodeIdInfo.manufactureDeviceName, var->v.valVIS_STR, SHORT_STRING_LEN );  
      Lg << DBG << "manufacture device name: {" <<canNodeIdInfo.manufactureDeviceName <<"}" << endl;
    }
  ode = od->seek( 0x1009, 0 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  strncpy( canNodeIdInfo.hardwareVersion, var->v.valVIS_STR, SHORT_STRING_LEN ); 
      Lg << DBG << "hardware version: {" <<canNodeIdInfo.hardwareVersion <<"}" << endl;
    }
  ode = od->seek( 0x100A, 0 );
  if( ode != NULL ) 
    {
      var = &(ode->currentValue);
      if( var->isSet() )  strncpy( canNodeIdInfo.softwareVersion, var->v.valVIS_STR, SHORT_STRING_LEN );
      Lg << DBG << "software version: {" <<canNodeIdInfo.softwareVersion <<"}" << endl; 
    }
  canNodeIdInfo.nodeId = nodeId;
  
  return SEQUENCER_NO_ERROR;
}

int CanNodeImage::seqExportPdoVariables(int *callNo, void *a)
{
  // register PDO handlers  

  for( int i = 0; i < MAX_PDO; i++ )
  {
      Lg << DBG << "TPDO_" << i +1 << " COBID = 0x" << hex << pdoClient->getTpdoCobId(i) << dec << endl;      
      if( pdoClient->getTpdoCobId(i) != 0 )  canMessageDispatcher->addMessageHandler(pdoClient->getTpdoCobId(i), this, &CanNodeImage::pdoHandler,i );
  }
  for( int i = 0; i < MAX_PDO; i++ )
      Lg << DBG << "RPDO_" << i +1 << " COBID = 0x" << hex << pdoClient->getRpdoCobId(i) << dec << endl;
  
  buildPropertiesSet();
  server->assignPropertyList(this);
  Lg << INFO << "PDO CONFIGURATION FINISHED FOR NODE 0x" << hex << nodeId << dec << endl;  
  return SEQUENCER_NO_ERROR;
}



int CanNodeImage::seqSetCurrentValuesForRpdoMappedVariables(int *callNo, void *a )
{
  
    if( (*callNo) > 0 )  
    {
        // read-write response check
        if( checkSDOwrite() != SEQUENCER_NO_ERROR ) return SEQUENCER_CONTINUE;  
    }
      
    int nItems =  pdoClient->getRpdoVariablesCount();
    if( nItems <= *callNo ) return SEQUENCER_NO_ERROR;
    MappedVariable * mv = pdoClient->getRmappedVariable( *callNo );
     ++(*callNo);
    ODentry * ode = od->seek(mv->index, mv->subindex);
    Lg << DBG << "seqSetDefaultValuesForRpdoMappedVariables:  submiting read for index = 0x" << hex << mv->index << ", subindex = 0x" << mv->subindex << ", value = 0x" << ode->getDefaultValuePtr()->valINT32 << endl;
    if( ode != NULL ) return submitSDOread(mv->index,mv->subindex, ode->getCurrentValuePtr(), ode->dataTypeCode );
    return SEQUENCER_ABORT;
}
  
ODentry * CanNodeImage::findNextWritableIndex(int currentIndex, int indexLimit, int subindex)
{
  ODentry * ode;
  while( ++currentIndex <= indexLimit )
  {
      ode = od->seek( currentIndex, subindex );
      if( ode != NULL )
      {
          if( ode->accessTypeCode & WRITE ) return ode;
      }
  }
  return NULL;   // no writable OD entry for given index range and subindex
}

ODentry * CanNodeImage::findNextWritableIndexSubindex(OD_WRITE_SEQUENCE_PAR *p)
{     
    while( 1 )
    { 
        ++p->startSubindex;
        p->ode = od->seek( p->startIndex, p->startSubindex );
        if( p->ode == NULL || !(p->ode->accessTypeCode & WRITE) )
        {
            if( ++p->startSubindex <= p->endSubindex ) continue;
           
            if( ++p->startIndex <= p->endIndex )
            {
                p->startSubindex = p->skipSubindex;
                continue;
            }
            p->ode = NULL;
        }       
        return p->ode;        
    }
    
    return NULL;   
}

//-----------------------------------[TINE PROPERTIES]----------------------------------------
int CanNodeImage::property_IdInfo( char *,  TINE_ARGS )
{
 if( dout->dFormat != CF_STRUCT ) return illegal_format;
 if( strcmp( dout->dTag, "NODEIDINFO")) return illegal_format;
  if( dout->dArrayLength < 1 ) //sizeof( CanNodeIdInfo ) )
    {
      Lg << DBG << "CanNodeIdInfohandler: wrong data size = " <<  dout->dArrayLength <<", expected = " << sizeof( CanNodeIdInfo)  << endl;
      return illegal_data_size;
    }
  memcpy( dout->data.vptr, &canNodeIdInfo, sizeof( CanNodeIdInfo ) ); 
  return 0;
}

int CanNodeImage::property_NMState( char *,  TINE_ARGS )
{
 if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG  ) return illegal_format;
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      *(dout->data.lptr) = heartBeatConsumer->getNodeState();
    }
  if( access & CA_WRITE )
    {    
      int nmtRequest;
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_LONG  ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;     
      switch( nmtRequest = *(din->data.lptr))
      {
        case START_NODE:
        case STOP_NODE:  
        case ENTER_PREOPERATIONAL:
        case RESET_NODE:
        case RESET_COMMUNICATION:
          sendNMT( nmtRequest );
          break;
      default:
        return invalid_data;
        break;
      }
    }
  return 0;
}

int CanNodeImage::property_nodeGuardState( char *, TINE_ARGS )
{ 
  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG  ) return illegal_format;
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      *(dout->data.lptr) = ( nodeGuardProducer->isEnabled() ? 1 : 0 );
    }
  if( access & CA_WRITE )
    {     
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_LONG  ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;     
      if(*(din->data.lptr)) nodeGuardProducer->enable(); 
      else  nodeGuardProducer->disable(); 
    }
  return 0;
}

int CanNodeImage::property_nodeGuardPeriod(char *,  TINE_ARGS )
{
  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG  ) return illegal_format;
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      *(dout->data.lptr) = nodeGuardProducer->getInterval();
    }
  if( access & CA_WRITE )
    {    
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_LONG  ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;     
      if( *(din->data.lptr) < 1 ) return invalid_data;      
      nodeGuardProducer->setInterval(*(din->data.lptr)); 
    }
  return 0;
}

int CanNodeImage::property_hbeatControl(char *,  TINE_ARGS )
{
  int errCode; 
  ODentry * ode;
  int hbTime;
  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG  ) return illegal_format;
      if( dout->dArrayLength < 2 ) return illegal_data_size;
      if( nodeGuardNeeded )
        {
          dout->data.lptr[0] = nodeGuardProducer->isEnabled();
          dout->data.lptr[1] = nodeGuardProducer->getInterval();
        }
      else
        {
          ode = od->seek( 0x1017,0);
          if( ode == NULL ) hbTime = 0;
          else hbTime = ode->getCurrentValuePtr()->valINT16;
          dout->data.lptr[0] = hbTime != 0;
          dout->data.lptr[1] = hbTime;
        }
    }
  if( access & CA_WRITE )
    {    
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_LONG  ) return illegal_format;
      if( din->dArrayLength < 2 ) return illegal_data_size;     
      if( din->data.lptr[1] < 0 ) return invalid_data;  
      if( nodeGuardNeeded )
        {
          nodeGuardProducer->setEnable(din->data.lptr[0]); 
          nodeGuardProducer->setInterval(din->data.lptr[1]); 
        }
      else
        {
          if( sdoClient->available() )
            {
              errCode = sdoClient->write( 0x1017, 0, &(din->data.lptr[1]), CO_INT16 ); 
              return errCode == 0 ? not_ready : 600 + errCode;
            }
          if( sdoClient->done() )
            { 
              errCode = sdoClient->getError();   
              if( errCode == ERROR_ABORT ) setEqmCompletion( sdoClient->getAbortDescription() );
              else setEqmCompletion( sdoClient->getErrorDescription() );
              sdoClient->reset(); 
              if( errCode == NO_ERROR )
                {                  
                  ode = od->seek(0x1017,0);
                  if( ode != NULL ) ode->assignCurrentValue((char*)( &(din->data.lptr[1])));
                  heartBeatConsumer->setPeriod( 2 * din->data.lptr[1] );
                  return 0;
                }
              return 600 + errCode;
            }
          else  return not_ready; 
        }
    }
  return 0; 
}

int CanNodeImage::property_stateStat(char *,  TINE_ARGS )
{
  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG  ) return illegal_format;
      if( dout->dArrayLength < 10 ) return illegal_data_size;
       
      memcpy( dout->data.lptr, heartBeatConsumer->getStateCounters(), 4 * sizeof(long) ); 
      memcpy( dout->data.lptr+4, heartBeatConsumer->getStateEntryTimes(), 5 * sizeof(long) );
      dout->data.lptr[9] = heartBeatConsumer->getNodeState();
      return 0;
    }

  return illegal_read_write;
}

int CanNodeImage::property_tpdoVariables( char *, TINE_ARGS )
{
  if( access & CA_READ )
    {      
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_STRUCT  ) return illegal_format;
      if( strcmp( dout->dTag, "MAPPEDVAR") ) return illegal_format;
      if( dout->dArrayLength < MAX_PDO * 8 ) return illegal_data_size;
      memcpy( dout->data.vptr,TPDOvariables, sizeof( TPDOvariables ) ); 
      return 0;
    }
  return illegal_read_write;
} 

int CanNodeImage::property_rpdoVariables( char *, TINE_ARGS )
{
  if( access & CA_READ )
    {      
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_STRUCT  ) return illegal_format;
      if( strcmp( dout->dTag, "MAPPEDVAR") ) return illegal_format;
       if( dout->dArrayLength < MAX_PDO * 8 ) return illegal_data_size;
      memcpy( dout->data.vptr,RPDOvariables, sizeof( RPDOvariables ) ); 
      return 0;
    }
  return illegal_read_write;
}

int CanNodeImage::property_odEntry( char *, TINE_ARGS )
{
  int errCode;
  int idx;
  int subidx;
  int dataType;
  unsigned int size;
  char *data;
  ODentryDataInfo *p;
  ODentry * ode;
  
  int lastCall = (access & CA_LAST) && !(access & CA_FIRST);
  int newSDOchannelAccessRequest = ((access & CA_REPEAT) == 0 && !lastCall);  // CA_REPEAT is set only if the handler is called as result of not_ready return code
  
  if( access & CA_WRITE )
    {       
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_STRUCT  && din->dFormat != CF_LONG && din->dFormat != CF_TEXT ) return illegal_format;
      if( din->dFormat == CF_STRUCT )
      {
	if( strcmp( din->dTag, "ODENTRYINFO") ) return illegal_format;
	if( din->dArrayLength < 1 ) return illegal_data_size;  
	p = (ODentryDataInfo*)din->data.cptr;   
	idx = p->index;
	subidx = p->subindex;
	dataType = p->dataCanType;
	size = p->dataSize;
	data = p->data;
      }
      else
      {
	  if( din->dFormat == CF_LONG )
	  {
	    if( din->dArrayLength < 3 ) return illegal_data_size;  	  	  
	    idx = din->data.lptr[0]; 
	    subidx = din->data.lptr[1];
	    data = (char*)&din->data.lptr[2];	    
	  }
	  else
	  {
	      // CF_TEXT
	      char * words[3];
	      int nWords = crack(din->data.cptr ,words," ,", 3);	      
	      if( nWords < 3 ) return illegal_data_size;
	      idx = strtol(words[0],NULL,0);
	      subidx = strtol(words[1], NULL, 0);
	      data = words[2];
	  }
	 
	  dataType = CO_DATA_TYPE_UNKNOWN;
	  size = 0;  // unknown requested data size
      }
     
      if( newSDOchannelAccessRequest )
      {
	  if( sdoClient->available() )
	    {          
	      if( ( ode = od->seek( idx, subidx ) ) == NULL )
		{
		  // no such entry, try obtain type from din              
		  if( (dataType ==  CO_DATA_TYPE_UNKNOWN) || ( dataType == CO_NO_DATA ))
		    {
		      dataType = CO_UINT32; // no entry in OD, no type indication... make a best guess
		      if( size == 0 ) sdoClient->setDataSize(ODentry::getDataTypeSize(dataType));
		      else sdoClient->setDataSize( size );  
		      Lg << WARNING << "--Warning: Node 0x" << hex << getNodeId() << " request to write idx = 0x" 
			 << idx <<", subidx 0x" <<  subidx 
			 <<", but no such entry in the local OD."
			 << dec << endl;  
		    }
		}
	      else dataType = ode->dataTypeCode;        
	      if(din->dFormat != CF_LONG) convertStringToCanData( data, data, dataType ); //CF_TEXT and CF_STRUCT pass data as string

	      errCode = sdoClient->write( idx, subidx, data, dataType );  
	      if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode ) 
					    <<", 0x" << errCode << dec << endl; 
	      return errCode == 0 ? not_ready : 600 + errCode;
	    }
	  else return operation_busy;  // bail out, somebody else is using the SDOchannel
      }	  
      
      if( sdoClient->done() )
        { 
          int errCode = sdoClient->getError();   
          if( errCode == ERROR_ABORT ) setEqmCompletion( (char*) sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset(); 
          return errCode ? 600 + errCode : 0;
        }
      else  // still waiting for the ack from the CAN node
      {
	  if( lastCall )  
	  {
	      sdoClient->reset();
	      return operation_timeout;
	  }
	  return not_ready;  // call me again
      } 
    }

  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( (dout->dFormat != CF_STRUCT) && (dout->dFormat != CF_LONG) && (dout->dFormat != CF_TEXT)) return illegal_format;  
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      if( din->dFormat == CF_LONG  ) 
        {
          if( din->dArrayLength < 1 ) return illegal_data_size;  	  
	  if( din->dArrayLength ==  1) subidx = 0;
	  else subidx = din->data.lptr[1];
	  size = 512;	  
	  idx = din->data.lptr[0]; 	  
        }
      else
        {
          if( din->dFormat != CF_STRUCT ) return illegal_format; 
          if( strcmp( din->dTag, "ODENTRYINFO") ) return illegal_format;
          if( din->dArrayLength < 1 ) return illegal_data_size;   
	  p = (ODentryDataInfo*)din->data.cptr;
          idx = p->index;
	  subidx = p->subindex;
	  size = p->dataSize;
        }
      if( newSDOchannelAccessRequest )
      {
	  if( sdoClient->available() )
	    {          
	      if((errCode = sdoClient->read( idx, subidx, size )) != NO_ERROR ) 
		Lg << ERROR << "--Error: node = 0x" << hex << getNodeId() << ", SDOclient->read returned error: " << sdoClient->getErrorDescription( errCode ) <<", 0x" << errCode << dec << endl; 
	      if( errCode == ERROR_ABORT ) setEqmCompletion( sdoClient->getAbortDescription() );
	      else setEqmCompletion( sdoClient->getErrorDescription() );
	      return errCode == 0 ? not_ready : 600 + errCode;
	    }
	  else return operation_busy;
      }
      if( sdoClient->done() )
        { 
          if( (errCode = sdoClient->getError()) == NO_ERROR )
            {             
              size = sdoClient->getDataSize();              
              int dataCanType;
              size = (size > 512) ? 512 : size;  //in case of domain data or strings copy only 512 bytes 
      
              ode = od->seek( idx, subidx );
              if( ode == NULL ) 
                {
                  Lg << ERROR << "-Warning: node = 0x" << hex << getNodeId() << ", property ODentry, no such index, subindex in local Object Dictionary0x"
                     << idx << ", 0x" << subidx << dec << endl;
                  dataCanType = CO_DATA_TYPE_UNKNOWN;
                }
              else dataCanType = ode->dataTypeCode;              
              if( dout->dFormat == CF_STRUCT )
	      { 
		p = (ODentryDataInfo*)dout->data.cptr;
		p->index = idx;
		p->subindex = subidx;
		p->dataCanType = dataCanType;	      
		p->dataSize = size;
		memcpy(p->data, sdoClient->getBuffer(), size );             
	      }
	      else
	      {
		  if( dout->dFormat == CF_TEXT )
		  {	     
		     strncpy(dout->data.cptr, sdoBufferToString(sdoClient->getBuffer(), dataCanType, size), dout->dArrayLength );
		  }
		  else
		  {
		    if( size > sizeof(long)) size = sizeof(long);
		    memcpy(dout->data.lptr, sdoClient->getBuffer(), size);
		  }
	      }
              sdoClient->reset();             
              return 0;   
            }
          if( errCode == ERROR_ABORT ) setEqmCompletion( sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset(); 
          return 600 + errCode;
        }
      else  // still waiting for the data from the CAN node
      {
	  if( lastCall ) 
	  {
	      sdoClient->reset();	  
	      return operation_timeout;
	  }
	  return not_ready;  // call me again
      } 
    }
  return 0;
}

int CanNodeImage::property_odEntryLong( char *, TINE_ARGS )
{
  int errCode;

  if( access & CA_WRITE )
    {       
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_LONG  ) return illegal_format;
      if( din->dArrayLength < 3 ) return illegal_data_size;     
      if( sdoClient->available() )
        {
          int idx = din->data.lptr[0]; 
          int subidx = din->data.lptr[1];
          if( din->dArrayLength == 4 ) 
            {
              //data size indicated
              sdoClient->setDataSize( din->data.lptr[3] );
              errCode = sdoClient->write( idx, subidx, &(din->data.lptr[2]), CO_DATA_TYPE_UNKNOWN );
            }
          else errCode = sdoClient->write( idx, subidx, &(din->data.lptr[2]), CO_INT32 );  // any other CO_type here?
          if( errCode != NO_ERROR ) 
            Lg << ERROR << "--Error: node 0x" << hex << getNodeId() << ", SDOclient->read returned error: " << sdoClient->getErrorDescription( errCode ) <<", 0x" << errCode << dec << endl; 
          return errCode == 0 ? not_ready : 600 + errCode;
        }
      if( sdoClient->done() )
        { 
          int errCode = sdoClient->getError();   
          if( errCode == ERROR_ABORT ) setEqmCompletion( (char*) sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset(); 
          return errCode ? 600 + errCode : 0;
        }
      else  return not_ready; 
    }

  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG  ) return illegal_format;  
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      if( din->dFormat != CF_LONG  ) return illegal_format;
      if( din->dArrayLength < 2 ) return illegal_data_size;
      if( sdoClient->available() )
        {
                        
          int idx = din->data.lptr[0]; 
          int subidx = din->data.lptr[1];
          int size = 512;
          if( dout->dArrayLength == 3 ) size = dout->data.lptr[2]; 
          if((errCode = sdoClient->read( idx, subidx, size )) != NO_ERROR ) 
            Lg << ERROR << "--Error: Node 0x" << hex << getNodeId() << ", SDOclient->read returned error: " << sdoClient->getErrorDescription( errCode ) <<", 0x" << errCode << dec << endl; 
          if( errCode == ERROR_ABORT ) setEqmCompletion( sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          return errCode == 0 ? not_ready : 600 + errCode;
        }
      if( sdoClient->done() )
        { 
          if( (errCode = sdoClient->getError()) == NO_ERROR )
            { 
              //dumpSDOdataRead();  
              int size = sdoClient->getDataSize();
              size = (size > 4) ? 4 : size;  //in case of domain data or strings copy only 4 bytes
              memcpy(dout->data.cptr, sdoClient->getBuffer(), size );             
              sdoClient->reset();             
              return 0;   
            }
          if( errCode == ERROR_ABORT ) setEqmCompletion( sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset(); 
          return 600 + errCode;
        }
      else  return not_ready; 
    }
 
 
  return 0;
}


int CanNodeImage::property_firmwareUpgrade( char *, TINE_ARGS )
{
  int errCode;
  int lastCall = (access & CA_LAST) && !(access & CA_FIRST);
  int programmingRequest = ((access & CA_REPEAT) == 0 && !lastCall);  // CA_REPEAT is set only if the handler is called as result of not_ready return code

  if( access & CA_WRITE )
    {
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_TEXT ) return illegal_format;
      if( din->dArrayLength > 256 ) return illegal_data_size;
      if( programmingRequest )
      {
	  if( flashProgrammer->available() )
	    {
              char fileName[256];
              strcpy( fileName, din->data.cptr );
              Lg << INFO << "request to start programmer for node: " << getName() << ", file name: " << fileName << endl;
	      errCode = flashProgrammer->start( fileName );              	      
	      return errCode == 0 ? 0 : 600 + errCode;
              //return errCode == 0 ? not_ready : 600 + errCode;
	    }
          else return operation_busy;  // bail out, programmer not available
      }
      
    }

  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG) return illegal_format;
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      dout->data.lptr[0] = flashProgrammer->getProgress();
    }
  return 0;
}

int CanNodeImage::property_firmwareUpgradeStatus( char *, TINE_ARGS )
{
    if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_INT32  ) return illegal_format;
      if( dout->dArrayLength < 4 ) return illegal_data_size;

      dout->data.lptr[0] = flashProgrammer->getStatus();
      dout->data.lptr[1] = flashProgrammer->getFileParity();
      dout->data.lptr[2] = flashProgrammer->getRamParity();
      dout->data.lptr[3] = flashProgrammer->getFlashParity();
      return 0;
    }
  return illegal_read_write;
}

int CanNodeImage::property_firmwareUpgradeEraseAndReprogram( char *, TINE_ARGS )
{
  int errCode;
 
  if( access & CA_WRITE )
    {
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_INT32 ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;
      if( din->data.lptr[0] == 1 )
      {
          if( flashProgrammer->available() )
          {
              errCode = flashProgrammer->unlockForErasingAndProgramming();
              return errCode == 0 ? 0 : errCode + 600;
          }
          else return operation_busy;  // bail out, programmer not available
      }
    }
            
  return 0; 
}


int CanNodeImage::property_firmwareUpgradeGetControlSums( char *, TINE_ARGS )
{
  int errCode;
  int lastCall = (access & CA_LAST) && !(access & CA_FIRST);
  int csRequest = ((access & CA_REPEAT) == 0 && !lastCall);  // CA_REPEAT is set only if the handler is called as result of not_ready return code

  if( access & CA_WRITE )
    {
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_INT32 ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;
      if( dout->dFormat != CF_INT32 ) return illegal_format;
      if( dout->dArrayLength < 2 ) return illegal_data_size;

      if( csRequest && din->data.lptr[0] == 1 )
      {
	  if( sdoClient->available() )
	    {
              int req = 32;
              errCode = sdoClient->write( 0x3300,2,&req,CO_INT32 );
	      if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )
					    <<", 0x" << errCode << dec << endl;
	      return errCode == 0 ? not_ready : 600 + errCode;
	    }
          else return operation_busy;  // bail out, programmer not available
      }

      if( sdoClient->done() )
        {
          int errCode = sdoClient->getError();
          if( errCode == ERROR_ABORT ) setEqmCompletion( (char*) sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset();
          if( errCode ) return 600 + errCode;
          if( sdoClient->getIndex() != 0x3301 )
          {
              errCode = sdoClient->read(0x3301,1,flashProgrammer->getRamParityPtr(), CO_INT32);
	      if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->read returned error: " << sdoClient->getErrorDescription( errCode )
					    <<", 0x" << errCode << dec << endl;
	      return errCode == 0 ? not_ready : 600 + errCode;
          }
          else
          {
            if(  sdoClient->getSubindex() == 2 )
            {
                dout->data.lptr[0] = flashProgrammer->getRamParity();
                dout->data.lptr[1] = flashProgrammer->getFlashParity();
                return 0;  // all done
            }
            errCode = sdoClient->read(0x3301,2,flashProgrammer->getFlashParityPtr(), CO_INT32);
            if( errCode != NO_ERROR )
            {
              Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )  <<", 0x" << errCode << dec << endl;
              sdoClient->reset();
              return errCode + 600;
            }
          }          
        }
	return not_ready;  // call me again
    }

  return 0;
}


int CanNodeImage::property_firmwareUpgradeSendData( char *, TINE_ARGS )
{
  int errCode;
 
  if( access & CA_WRITE )
    {
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_TEXT ) return illegal_format;
      if( din->dArrayLength > 256 ) return illegal_data_size;
      if( flashProgrammer->available() )
        {
          char fileName[256];
          strcpy( fileName, din->data.cptr );
          Lg << INFO << "request to send reprogramming data for node: " << getName() << ", file name: " << fileName << endl;
          errCode = flashProgrammer->sendData( fileName );
          if( errCode != NO_ERROR )
          {              
              Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )
                                        <<", 0x" << errCode << dec << endl;
              sdoClient->reset();
              return 600 + errCode;
          }
          return 0;
        }
      else return operation_busy;  // bail out, programmer not available
    }

  if( access & CA_READ )
    {
      if( dout == NULL ) return illegal_data_size ;
      if( dout->dFormat != CF_LONG) return illegal_format;
      if( dout->dArrayLength < 1 ) return illegal_data_size;
      dout->data.lptr[0] = sdoClient->getProgress();
    }
  return 0;
}

int CanNodeImage::property_firmwareUpgradeSwitchToMainProgram( char *, TINE_ARGS )
{
  int errCode;
  int lastCall = (access & CA_LAST) && !(access & CA_FIRST);
  int switchRequest = ((access & CA_REPEAT) == 0 && !lastCall);  // CA_REPEAT is set only if the handler is called as result of not_ready return code

  if( access & CA_WRITE )
    {
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_INT32 ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;
      if( switchRequest && din->data.lptr[0] == 1 )
      {
	  if( sdoClient->available() )
	    {
              int key = 0x12345678;
              errCode = sdoClient->write(0x3300,1,&key,CO_INT32);             
	      if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )
					    <<", 0x" << errCode << dec << endl;
	      return errCode == 0 ? not_ready : 600 + errCode;              
	    }
          else return operation_busy;  // bail out, programmer not available
      }
      
      if( sdoClient->done() )
        {
          int errCode = sdoClient->getError();
          if( errCode == ERROR_ABORT ) setEqmCompletion( (char*) sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset();
          if( errCode ) return 600 + errCode;
          int mainProgJump = 1;          
          errCode = sdoClient->write(0x3300,2,&mainProgJump,CO_INT32);           
	  if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )  <<", 0x" << errCode << dec << endl;
          sdoClient->reset();
          return errCode ? 600 + errCode : 0;
        }      
	return not_ready;  // call me again      
    }
  
  return 0;
}

int CanNodeImage::property_firmwareUpgradeSwitchToProgrammingMode( char *, TINE_ARGS )
{
  int errCode;
  int lastCall = (access & CA_LAST) && !(access & CA_FIRST);
  int switchRequest = ((access & CA_REPEAT) == 0 && !lastCall);  // CA_REPEAT is set only if the handler is called as result of not_ready return code

  if( access & CA_WRITE )
    {
      if( din == NULL ) return illegal_data_size ;
      if( din->dFormat != CF_INT32 ) return illegal_format;
      if( din->dArrayLength < 1 ) return illegal_data_size;
      if( switchRequest && din->data.lptr[0] == 1 )
      {
	  if( sdoClient->available() )
	    {
              int key = 0x12345678;
              errCode = sdoClient->write(0x3300,1,&key,CO_INT32);             
	      if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )
					    <<", 0x" << errCode << dec << endl;
	      return errCode == 0 ? not_ready : 600 + errCode;              
	    }
          else return operation_busy;  // bail out, programmer not available
      }
      
      if( sdoClient->done() )
        {
          int errCode = sdoClient->getError();
          if( errCode == ERROR_ABORT ) setEqmCompletion( (char*) sdoClient->getAbortDescription() );
          else setEqmCompletion( sdoClient->getErrorDescription() );
          sdoClient->reset();
          if( errCode ) return 600 + errCode;
          int brennProgJump = 2;
          errCode = sdoClient->write(0x3300,2,&brennProgJump,CO_INT32);             
	  if( errCode != NO_ERROR )  Lg << ERROR << "--Error: node 0x" << hex << nodeId << ", SDOclient->write returned error: " << sdoClient->getErrorDescription( errCode )  <<", 0x" << errCode << dec << endl;
          sdoClient->reset();
          return errCode ? 600 + errCode : 0;
        }      
	return not_ready;  // call me again      
    }
  
  return 0;
}
     

int CanNodeImage::property_emcy( char *, TINE_ARGS )
{
    if( access & CA_WRITE ) return illegal_read_write;    
    if( dout == NULL ) return illegal_data_size ;
    if( dout->dFormat != CF_STRUCT  ) return illegal_format;
    if( strcmp( dout->dTag, "EMCY" ) ) return illegal_format;
    if( dout->dArrayLength < 1 ) return illegal_data_size;
    if( din->dFormat != CF_LONG  ) return illegal_format;
    if( din->dArrayLength < 2 ) return illegal_data_size;
    
    long long ts = (long long)(din->data.lptr[0]) * 1000 + din->data.lptr[1];    
    int idx = emcyList.seek(ts);    
    
    EMCY_INFO  * e = (EMCY_INFO*)dout->data.vptr;
    if( idx < 0 )
    {
        e->isMessage = 0;        
        return 0;
    }    
    emcyList.store(idx, e); 
    return 0;
}

int CanNodeImage::firmwareUpgrade()
{
    return flashProgrammer->isActive();
}

void CanNodeImage::convertStringToCanData( char * data, char *str,  int dataType )
{
  switch( dataType )
    {
    case CO_NO_DATA:
    case CO_DATA_TYPE_UNKNOWN:
      *data = '\0';
      break;
    case CO_BOOL1:
    case CO_INT8: 
    case CO_INT16: 
    case CO_INT32: 
    case CO_UINT8:      
    case CO_UINT16:      
    case CO_UINT32:  
    case CO_INT24: 
    case CO_UINT24:      
      long t = strtol( str, NULL, 0 );
      memcpy( data, &t, sizeof(t));
      break;       
    case CO_UINT40:  
    case CO_UINT48:  
    case CO_UINT56:    
    case CO_UINT64: 
    case CO_INT40:   
    case CO_INT48: 
    case CO_INT56:  
    case CO_INT64:  
      long long lt = strtoll( str, NULL, 0 );
      memcpy( data, &lt, sizeof(lt));
      break;
    case CO_REAL32: 
      float f = strtof(str, NULL); 
      memcpy( data, &f, sizeof(f));
      break;
    case CO_REAL64:
      double df = strtod(str, NULL );
      memcpy( data, &df, sizeof(f));
      break;   
    case CO_VIS_STR:
    case CO_OCTET_STR: 
    case CO_BIT_STR:
      strcpy( data, str );     
      break;
    case CO_UNICODE_STR:    
    case CO_TIME_OF_DAY:    
    case CO_TIME_DIFF:     
    case CO_DOMAIN:
      Lg << WARNING << "-Warning: node: 0x" << hex << getNodeId() << dec << ", unhandled data type: " << dataType << endl;
        *data = '\0';
      break;      
    }
}

void CanNodeImage::dump()
{
  Lg << DBG << endl << "Dump of : " << getName() << endl;
  Lg << DBG << "NodeId = 0x" << hex << getNodeId() << dec << ", baud rate = " << getBaudRate() << endl;
  Lg << DBG << "Need initialization : " << needInitialization << endl;
  Lg << DBG << "NMT state = 0x" << hex << heartBeatConsumer->getNodeState() << dec << endl;
  Lg << DBG << "SDOclient: done = " << sdoClient->done() << ", available = " << sdoClient->available() << endl;
  Lg << DBG << endl;
}

void CanNodeImage::dump(int pathDepth, char ** devicePath )
{
  if( pathDepth == 1 ) 
    {
      if( !strcmp( *devicePath, "od" ) ) 
        {
          od->dump();
          return;
        }
      if( !strcmp( *devicePath, "sdo" ) ) 
        {
          dumpSDOdataRead();
          return;
        }
    }

  Lg << DBG << "unhandled request:" << *devicePath++;
  for( int i = 0; i < pathDepth; i++ ) Lg << DBG << "\\" << *devicePath;
  Lg << DBG << endl;
    
}

void CanNodeImage::ignoreBootMessages(int ignorePeriodMs)
{
    if( ignorePeriodMs == 0 ) ignoreBootup = 0;
    else
    {
        ignoreBootup = 1;
        ignoreBootupTimer->start( ignorePeriodMs );
    }
}

EmcyList::EmcyList()
{
    clear();
}

void EmcyList::clear()
{
    memset( emcyItems, '\0', sizeof(emcyItems));
}
void EmcyList::add( short emcyCode, char errorRegister, char manufactureSpecCode[5] )
{
   // next EMCY message is adedd always to the first element of the array
    memmove( &emcyItems[1], &emcyItems[0], sizeof(EMCY_ITEM) * (MAX_EMCY_ITEMS -1) );         
    emcyItems[0].timeStamp = HrTimer::hrtimer();
    emcyItems[0].emcyCode = emcyCode;
    emcyItems[0].errorRegister = errorRegister;
    memcpy( emcyItems[0].manufactureSpecificCode, manufactureSpecCode, 5 );
}

int EmcyList::seek( long long timeStamp )
{
   // returns -1, if not found   

   int i = 0;
   for( i = 0; i < MAX_EMCY_ITEMS; i++ ) 
   {      
       if( emcyItems[i].timeStamp < timeStamp ) return --i;
   }
   return --i;
}

void EmcyList::store(int idx, EMCY_INFO * e)
{
    e->isMessage = 1;
    e->emcyCode = emcyItems[idx].emcyCode;
    e->errorRegister = emcyItems[idx].errorRegister;
    memcpy( e->manufactureSpecificCode, emcyItems[idx].manufactureSpecificCode, 5 );
    strcpy( e->emcyText, CanNodeImage::getEmcyText(e->emcyCode ));
    strcpy( e->errorRegisterText, CanNodeImage::getErrorRegisterText(e->errorRegister));
    e->timeStamp[0] = emcyItems[idx].timeStamp / 1000;
    e->timeStamp[1] = emcyItems[idx].timeStamp % 1000;
}

