#ifndef __CAN_NODE_IMAGE_H__
#define __CAN_NODE_IMAGE_H__

#include "loggerClass/log.h"
#include "timeSlice.h"
#include "canMessage.h"
#include "edsParser.h"
#include "objectDictionary.h"
#include "sdo.h"
#include "sdoForwarder.h"
#include "pdo.h"
#include "nodeGuard.h"
#include "canMessageCollector.h"
#include "hrtimer.h"
#include "sequencer.h"
#include "heartBeat.h"
#include "tineServer.h"
#include "flashProgrammer.h"



extern Log Lg;

const int INIT_NO_ERROR = 0;
const int INIT_WRONG_DEVICE_TYPE = 1;
const int INIT_OD_NODE_MISMATCH = 2;
const int INIT_SDO_SUBMIT_ERROR = 4;
const int INIT_SDO_RW_ERROR = 8;
const int INIT_INTERNAL_ERROR = 16;

typedef struct
{
  long long timeStamp;
  short emcyCode;
  char  errorRegister;
  char  manufactureSpecificCode[5];   
} EMCY_ITEM;

typedef struct
{
  long  timeStamp[2];
  short emcyCode;
  char  errorRegister;
  char  manufactureSpecificCode[5];   
  char  isMessage;
  char  errorRegisterText[207];
  char  emcyText[48];          
} EMCY_INFO;

#define MAX_EMCY_ITEMS 16
class EmcyList
{
    private:
        EMCY_ITEM emcyItems[MAX_EMCY_ITEMS];
        int nItems;
        void clear();
    public:
        EmcyList();
        void add( short emcyCode, char errorRegister, char manufactureSpecCode[5] );
        void store(int idx, EMCY_INFO * e);
        int seek(long long timeStamp);        
};


const int CanToTineFormat[] =
  {
    CF_NULL,
    CF_BIT, 
    CF_BYTE, 
    CF_SHORT, 
    CF_INT32,
    CF_INT8, /* CF_UINT8, */
    CF_INT16, /* CF_UINT16, */
    CF_INT32, /*CF_UINT32, */
    CF_FLOAT,
    CF_TEXT,    /* CO_VIS_STR */
    CF_TEXT,    /* CO_OCTET_STR */
    CF_USTRING, /* CO_UNICODE_STR */
    CF_STRUCT,  /* CO_TIME_OF_DAY */
    CF_STRUCT,  /*  CO_TIME_DIFF */
    CF_TEXT,    /* CO_BIT_STR */
    CF_TEXT,    /*  CO_DOMAIN */
    CF_INT32, /* CF_UINT32, */  /* CO_INT24 */
    CF_DOUBLE,  /* CO_REAL64 */
    CF_LNGINT,  /* CO_INT40 */
    CF_LNGINT,  /* CO_INT48 */
    CF_LNGINT,  /* CO_INT56 */
    CF_LNGINT,  /* CO_INT64 */
    CF_INT32, /* CF_UINT32, */  /* CO_UINT24 */
    CF_LNGINT,  /* CO_UINT40 */
    CF_LNGINT,  /* CO_UINT48 */
    CF_LNGINT,  /* CO_UINT56 */
    CF_LNGINT /* CO_UINT64 */
  };

class PDOclient;
class CanMessageCollector;
class CanMessageDispatcher;
class FlashProgrammer;

typedef struct
{
  int startIndex;
  int startSubindex;
  int skipSubindex;
  int endIndex;
  int endSubindex;
  int value;
  int type;
  ODentry * ode;
} OD_WRITE_SEQUENCE_PAR;


const int SHORT_STRING_LEN = 32;


typedef struct 
{
 public:
  long nodeId;
  long deviceType;
  long vendorId;
  long productCode;
  long revisionNumber;
  long serialNumber;
  char manufactureDeviceName[SHORT_STRING_LEN+1];
  char dummy0[3];
  char hardwareVersion[SHORT_STRING_LEN+1];   
  char dummy1[3];
  char softwareVersion[SHORT_STRING_LEN+1];
  char dummy2[3];
} CanNodeIdInfo;


typedef struct
{
  short index;   
  short subindex;
  long long dataValue;    // any numeric value fits here  (64 bits of PDO)
  long long updateTime;   
  short dataSize;
  short dataFormat;
  short npdos;               // how many PDOs map this variable
  short pdos[MAX_PDO];       // which PDOs map this variable
  char variableName[64];
} MappedVariableValue;

typedef struct
{
  long index;
  long subindex;
  long dataCanType;
  long dataSize;
  char data[512];
} ODentryDataInfo;

typedef struct
{
  int nItems;
  int currentItem;
} DEFAULT_ENTRIES_READING;

const OD_ENTRY_IDX DefaultUpdateEntries[] = 
  { 
    {0x1000,0},
    {0x1008,0},
    {0x1009,0},
    {0x100A,0},
    {0x1018,1},
    {0x1018,2},
    {0x1018,3},
    {0x1018,4}
  };

class CanNodeImage:  public TineDevice
{
 private:
  unsigned int nodeId;
  unsigned int baudRate;
  unsigned int initializationErrorCode;
  int ignoreBootup;
  HrTimer * ignoreBootupTimer;
  FlashProgrammer * flashProgrammer;
  SDOclient * sdoClient;
  SDOforwarder * sdoForwarder;
  PDOclient * pdoClient;
  HeartBeatConsumer * heartBeatConsumer;
  NodeGuardProducer * nodeGuardProducer;
  EDSparser *EdsParser;
  ObjectDictionary *od;
  Sequencer * sequencer;
  CanMessageCollector * canMessageCollector;
  CanMessageDispatcher * canMessageDispatcher;
  CanNodeIdInfo canNodeIdInfo;
  EmcyList emcyList;
  ODentryText tempBuffEntry; 
  char buffer[2560];
  int needInitialization;
  short nodeGuardNeeded;
  int ready; //configured and can listen the hearbeats, switch internal states etc.
  MappedVariableValue TPDOvariables[MAX_PDO*8];
  MappedVariableValue RPDOvariables[MAX_PDO*8];
  struct ph
  {
    char name[33];
    int (CanNodeImage::*handler)( char *, TINE_ARGS );
  }  propertyHandlerList[MAX_PROPERTY];

  
  void construct(CanMessageDispatcher * cmd);
  void analyseOD();
  void storeTempBufferEntry();
  void sectionParsingHandler( char * sectionName, char * keyName, char *keyValue );
  void objectDictionaryParsingHandler( int index, int subindex, char * keyName, char *keyValue, int lineNumber );
  void convertStringToCanData( char * data, char *str,  int dataType );
  void configureRemoteNode();

  //---[sequencer step functions]----------------------------------------------------------------
  int seqReadDefaultEntries(int *callNo, void * a);
  int seqStoreDefaultEntriesForTine(int *callNo, void * a);
  int seqReadArraySize(int *callNo, void *a );
  int seqReadArrayBody(int *callNo, void *a );
  int submitSDOwrite( int index, int subindex, void *data, int dtype );
  int seqSetHeartBeatTime(int *callNo, void *a);
  int checkSDOwrite();
  int seqSetPdoParameters( int *callNo, void *a );
  int submitSDOread( int index, int subindex, int dtype );
  int submitSDOread( int index, int subindex, void *data, int dtype );
  int seqCheckDeviceType(int *callNo, void * a);
  int seqCheckNodeGuardNeeded(int *callNo, void * a); 
  int seqReadTpdoConfig(int *callNo, void * a);
  int seqReadTpdoMapping(int *callNo, void * a);  
  int seqReadRpdoMapping(int *callNo, void * a);
  int seqReadRpdoConfig(int *callNo, void * a);
  int seqExportPdoVariables(int *callNo, void * a); 
  int seqSetCurrentValuesForRpdoMappedVariables(int *callNo, void *a );
  // seqencer functions for dynamic PDO remapping
  int seqDisableRpdo(int *callNo, void * a);
  int seqDisableTpdo(int *callNo, void * a);
  int seqDisableRpdoMapping(int *callNo, void * a);
  int seqDisableTpdoMapping(int *callNo, void * a);
  int seqSetRpdoMapping(int *callNo, void * a);
  int seqSetTpdoMapping(int *callNo, void * a);
  int seqEnableRpdoMapping(int *callNo, void * a);
  int seqEnableTpdoMapping(int *callNo, void * a);
  int seqSetRpdoParam(int *callNo, void * a);
  int seqSetTpdoParam(int *callNo, void * a);
  int seqEnableRpdo(int *callNo, void * a);
  int seqEnableTpdo(int *callNo, void * a);
  // helpers for dynamic PDO remapping
  int continueDisablePdo( OD_WRITE_SEQUENCE_PAR *p );
  int continueEnablePdoMapping(int *callNo, OD_WRITE_SEQUENCE_PAR *p );
  int continueEnablePdo(int *callNo, OD_WRITE_SEQUENCE_PAR *p );
  int sendPdoConfig( int *callNo, OD_WRITE_SEQUENCE_PAR *p );

  void init();
  void bootup();
  void preoperational();
  void operational();
  void stopped();
  void addAutomaticProperty( MappedVariable * mv );
  int getAutomaticProperty( MappedVariable * mv, TineProperty * p );
  int canFormatToTineFormat( unsigned int canFormat );
  char * sdoBufferToString( char * data, int dataType, int dataSize );
  int defaultAutomaticPropertyHandler( char *property, TINE_ARGS );
  int property_NMState( char *, TINE_ARGS );
  int property_stateStat(char *,  TINE_ARGS );
  int property_nodeGuardState( char *, TINE_ARGS );  
  int property_nodeGuardPeriod( char *,  TINE_ARGS );
  int property_odEntry( char *, TINE_ARGS );
  int property_odEntryLong( char *, TINE_ARGS );
  int property_IdInfo( char *,  TINE_ARGS );
  int property_hbeatControl(char *,  TINE_ARGS );
  int property_tpdoVariables(char *,  TINE_ARGS );
  int property_rpdoVariables(char *,  TINE_ARGS );
  int property_emcy( char *, TINE_ARGS );
  int property_firmwareUpgrade( char *, TINE_ARGS );
  int property_firmwareUpgradeStatus( char *, TINE_ARGS );
  int property_firmwareUpgradeEraseAndReprogram( char *, TINE_ARGS );
  int property_firmwareUpgradeSendData( char *, TINE_ARGS );
  int property_firmwareUpgradeSwitchToMainProgram( char *, TINE_ARGS );
  int property_firmwareUpgradeSwitchToProgrammingMode( char *, TINE_ARGS );
  int property_firmwareUpgradeGetControlSums( char *, TINE_ARGS );

  void buildPropertiesSet();
  void addProperty( TineProperty * p, int (CanNodeImage::*h)(char * propertyName,  TINE_ARGS ));
  void addProperty( char * propName, DTYPE* out, DTYPE* in, int acc, char * dsc, int (CanNodeImage::*h)( char *, TINE_ARGS ), MappedVariable * mappedVariable = NULL);
  inline void setInitializationError( int e ) { initializationErrorCode = e; }
  ODentry * findNextWritableIndex(int currentIndex, int indexLimit, int subindex);
  ODentry * findNextWritableIndexSubindex(OD_WRITE_SEQUENCE_PAR *p);
  
 public: 
  CanNodeImage(CanMessageDispatcher * cmd = NULL, CanMessageCollector *cmc = NULL, char * dcfFile = NULL); 
  CanNodeImage(CanMessageDispatcher * cmd, CanMessageCollector * collector, char * dcfFile, char * nodeName, int nodeId);
  ~CanNodeImage();
  inline void connectMessageCollector( CanMessageCollector * collector )
    { 
      sdoClient->connectMessageCollector(canMessageCollector = collector);      
      pdoClient->connectMessageCollector(canMessageCollector); 
      nodeGuardProducer->connectMessageCollector(canMessageCollector);
    }
  inline void setCanMessageDispatcher( CanMessageDispatcher * cmd ) { canMessageDispatcher = cmd; }
  // static wrappers for callbacks
  static void sectionParsingHandlerWrapper( void * object, char * sectionName, char * keyName, char *keyValue ); 
  static void objectDictionaryParsingHandlerWrapper( void * object, int index, int subindex, char * keyName, char *keyValue, int lineNumber );

  void activate();
  int emcyHandler( int, CanMessage& m ); 
  int sdoHandler( int, CanMessage& m );
  int pdoHandler( int pdoNr, CanMessage& m );
  int nmtHandler(int, CanMessage& m );
  int parse( char * fileName );
  int parse( char * fileName, char *nodeName, int nodeID );
 
  static char * getErrorRegisterText( unsigned char c );
  static char * getEmcyText( unsigned short c );

  inline unsigned short getNodeId(void) { return nodeId; }
  void setNodeId( unsigned int id );
 
  inline int getBaudRate() { return baudRate; }

  inline void dumpOD() { od->dump(); }
  void dump();
  void dump( int pathDepth, char ** devicePath );

  void dumpSDOdataRead();
  void * getODcurrentValuePtr( int index, int subindex );
 
  // TINE RPC:
  int propertyHandler( char *property, TINE_ARGS );

  //---------------------[ USER INTERFACE ]-----------------------
  int sendNMT( int state );

  inline ObjectDictionary * getOD() { return od; } 
  inline PDOclient * getPDOclient() { return pdoClient; }
  inline SDOclient * getSDOclient() { return sdoClient; }
  inline SDOforwarder * getSDOforwarder() { return sdoForwarder; }
  inline int getTpdoVariablesCount() { return pdoClient->getTpdoVariablesCount(); } 
  inline int getRpdoVariablesCount() { return pdoClient->getRpdoVariablesCount(); }
  inline MappedVariable * getTmappedVariable( int varNo ) { return pdoClient->getTmappedVariable(varNo);}  
  inline MappedVariable * getRmappedVariable( int varNo ) { return pdoClient->getRmappedVariable(varNo);}
  inline MappedVariable * getTmappedVariable( int index, int subindex ) { return pdoClient->getTmappedVariable( index, subindex ); }
  inline MappedVariable * getRmappedVariable( int index, int subindex ) { return pdoClient->getRmappedVariable( index, subindex ); }
  inline int getNodeState() { return heartBeatConsumer->getNodeState(); }
  inline int isInitialized() { return heartBeatConsumer->getNodeState() != INIT; }
  inline int getInitializationError() { return initializationErrorCode; }
  void ignoreBootMessages( int ign );
  int firmwareUpgrade();
};

#endif
