

#include "flashProgrammer.h"

FlashProgrammer::FlashProgrammer( CanNodeImage *cni )
{
    motorolaSrecordReader = NULL;
    ramParity = -1;
    flashParity = -2;
    fileParity = -3;
    owner = cni;
    sdoClient = cni->getSDOclient();
    autoResetTimer = new HrTimer();
    binaryDataBuffer = NULL;
    state = status = IDLE;
    Lg << INFO << "FlashProgrammer constructed for " << cni->getName() << endl;
}

FlashProgrammer::~FlashProgrammer()
{
    deleteBinaryDataBuffer();
    delete autoResetTimer;
}

int FlashProgrammer::available()
{
    return state == IDLE && sdoClient->available();
}

int FlashProgrammer::start( char * fileName )
{
    return start( fileName, UNLOCK_BRENNMODE );
}

int FlashProgrammer::start( char * fileName, int initialState )
{
    int errCode;

    Lg << INFO << "FlashProgrammer started for node: " << owner->getNodeId() << endl;
    binaryDataBuffer = new char[160*1024];  // 160kB
    motorolaSrecordReader = new MotorolaSrecordReader(fileName);
    if( (errCode = motorolaSrecordReader->convert(binaryDataBuffer, 160*1024 ))
            != MotorolaSrecordReader::OK )
    {
        Lg << ERROR << "MotorolaSrecordReader error: " << errCode << endl;
        reset();
        return FlashProgrammer::DATA_CORRUPTED;
    }
    dataSize = motorolaSrecordReader->getDataSize();
    fileParity = motorolaSrecordReader->getDataByteSum();
    delete motorolaSrecordReader;
    state = initialState;
    return 0;
}

int FlashProgrammer::sendData( char* fileName )
{
   int errCode;

    if( !sdoClient->available() ) return ERROR_SDO_BUSY;
    Lg << INFO << "FlashProgrammer: sending data for node: " << owner->getNodeId() << endl;

    status = SENDING_DATA;
    state = SINGLE_STEP_SEND_DATA;
    binaryDataBuffer = new char[160*1024];  // 160kB
    motorolaSrecordReader = new MotorolaSrecordReader(fileName);
    if( (errCode = motorolaSrecordReader->convert(binaryDataBuffer, 160*1024 )) != MotorolaSrecordReader::OK )
    {
        Lg << ERROR << "MotorolaSrecordReader error: " << errCode << endl;
        return FlashProgrammer::DATA_CORRUPTED;
    }

    dataSize = motorolaSrecordReader->getDataSize();
    fileParity = motorolaSrecordReader->getDataByteSum();
    delete motorolaSrecordReader;

    sdoClient->setDataSize(dataSize);
    return sdoClient->write(0x320A,0,binaryDataBuffer,CO_DOMAIN);
}

void FlashProgrammer::handleSingleStepSendData()
{
    if( sdoClient->done() )
    {
       int errCode = sdoClient->getError();
       if( errCode )
       {
         if( errCode == ERROR_ABORT ) Lg << ERROR << "--ERROR: node " << owner->getNodeId() << " sending data aborted: " << sdoClient->getAbortDescription() << endl;
         else Lg << ERROR << "--ERROR: node " << owner->getNodeId() << " sending data error: " << sdoClient->getErrorDescription() << endl;
         reset();
         status = IDLE;
         sdoClient->reset();
         return;
       }
       int dataSize = sdoClient->getDataSize();
       sdoClient->reset();
       errCode = sdoClient->write( 0x3200,0x0A,&dataSize, CO_INT32 );
       if( errCode == 0 ) state = SINGLE_STEP_SIZE_WRITTEN;
       else
       {
            reset();
            status = IDLE;
       }
    }    
}

void FlashProgrammer::handleSingleStepSizeWritten()
{
    if( sdoClient->done() )
    {
       int errCode = sdoClient->getError();
       if( errCode )
       {
         if( errCode == ERROR_ABORT ) Lg << ERROR << "--ERROR: node " << owner->getNodeId() << " setting size aborted: " << sdoClient->getAbortDescription() << endl;
         else Lg << ERROR << "--ERROR: node " << owner->getNodeId() << " setting size error: " << sdoClient->getErrorDescription() << endl;
       }
     reset();
     status = IDLE;
     sdoClient->reset();
    }
}

int FlashProgrammer::done()
{
    return state == FINISHED;
}
char * FlashProgrammer::getAbortDescription()
{
    return sdoClient->getAbortDescription();
}
char * FlashProgrammer::getErrorDescription()
{
    return sdoClient->getErrorDescription();
}
void FlashProgrammer::reset()
{
    deleteBinaryDataBuffer();
    state = IDLE;
    Lg << INFO << "FLASH PROGRAMMER RESET" << endl;
}

void FlashProgrammer::deleteBinaryDataBuffer()
{
    if( binaryDataBuffer ) delete binaryDataBuffer;
    binaryDataBuffer = NULL;
}

void FlashProgrammer::activate()
{
    switch( state )
    {
        case IDLE:  return;
            break;
        case UNLOCK_BRENNMODE: handleUnlockBrennMode();
            break;
        case UNLOCK_WAIT_FOR_RESPONSE: handleUnlockWaitForResponse();
            break;
        case WAIT_FOR_BOOT_UP_STATE: handleWaitForBootupState();
            break;
        case WAIT_FOR_PREOPERATIONAL_STATE: handleWaitForPreoperationalState();
            break;
        case INIT_DOWNLOAD: handleInitDownload();
            break;
        case INIT_WAIT_FOR_RESPONSE: handleInitWaitForResponse();
            break;
        case SEND_DATA: handleSendData();
            break;
        case SEND_DATA_WAIT_FOR_RESPONSE: handleSendDataWaitForResponse();
            break;
        case DATA_SENT: handleDataSent();
            break;
        case ERASE_AND_PROGRAMMING: handleEraseAndProgramming();
            break;
        case DATA_PROGRAMMED: handleDataProgrammed();
            break;
        case CHECK_PROGRAMMING_STATUS: handleCheckProgrammingStatus();
            break;
        case WAIT_FOR_STATUS: handleWaitForStatus();
            break;
        case CALCULATE_PARITY: handleCalculateParity();
            break;
        case WAIT_FOR_CALCULATE_PARITY: handleWaitForCalculateParity();
            break;
        case CHECK_RAM_PARITY: handleCheckRamParity();
            break;
        case WAIT_FOR_RAM_PARITY: handleWaitForRamParity();
            break;
        case CHECK_FOR_FLASH_PARITY: handleCheckFlashParity();
            break;
        case WAIT_FOR_FLASH_PARITY: handleWaitForFlashParity();
            break;
        case UNLOCK_MAIN_MODE: handleUnlockMainMode();
            break;
        case JUMP_TO_MAIN_MODE: handleJumpToMainMode();
            break;
        case SINGLE_STEP_SEND_DATA: handleSingleStepSendData();
            break;
        case SINGLE_STEP_SIZE_WRITTEN: handleSingleStepSizeWritten();
            break;
        case SINGLE_STEP_ERASE_AND_REPROGRAM: handleSingleStepEraseAndReprogram();
            break;
        case SINGLE_STEP_DATA_PROGRAMMED: handleSingleStepDataProgrammed();
            break;
        case SINGLE_STEP_CHECK_PROGRAMMING_STATUS: handleSingleStepCheckProgrammingStatus();
            break;
        case SINGLE_STEP_WAIT_FOR_STATUS: handleSingleStepWaitForStatus();
            break;

        case FINISHED:
            if( autoResetTimer->expired() ) reset();
            status = IDLE;
            break;
    }
}

 void FlashProgrammer::handleUnlockBrennMode()
 {
    status = SWITCHING_TO_PROGRAMMING_MODE;
    unlock(UNLOCK_WAIT_FOR_RESPONSE);
 }

void FlashProgrammer::handleUnlockWaitForResponse()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError()) return;
    
    owner->ignoreBootMessages(3000);  // during next 3s boot-up messages will be ignored

    int goToBrennMode = 0x02;
    sdoClient->write(0x3300,2,&goToBrennMode, CO_INT32);    
    state = WAIT_FOR_BOOT_UP_STATE;
    sdoClient->reset(); // there will be no response, since node will reboot
}

void FlashProgrammer::handleWaitForBootupState()
{
    if( owner->getNodeState() == BOOT_UP ) state = WAIT_FOR_PREOPERATIONAL_STATE;
}
void FlashProgrammer::handleWaitForPreoperationalState()
{
    if( owner->getNodeState() == PREOPERATIONAL ) state = INIT_DOWNLOAD;
}
void FlashProgrammer::handleInitDownload()
{    
    if( checkSdoAvailable( 0x3200, 0xA) ) return;
    
    status = SENDING_DATA;

    int err = sdoClient->write( 0x3200,0x0A,&dataSize, CO_INT32 );
    if( err == 0 ) state = INIT_WAIT_FOR_RESPONSE;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x3200, subidx = 1, error = " << err << endl;
        reset();
    }

}

void FlashProgrammer::handleInitWaitForResponse()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    state = SEND_DATA;
}

void FlashProgrammer::handleSendData()
{
    Lg << INFO << "FlashProgrammer: sending program code for node: " << owner->getNodeId() << endl;
    if( checkSdoAvailable(0x320A, 0)) return;
    
    sdoClient->setDataSize(dataSize);
    int err = sdoClient->write( 0x320A,0,binaryDataBuffer, CO_DOMAIN );
    if( err == 0 ) state =  SEND_DATA_WAIT_FOR_RESPONSE;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x320A, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleSendDataWaitForResponse()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    Lg << INFO << "FlashProgrammer: sent program code for node: " << owner->getNodeId() << endl;
    status = REPROGRAMMING;
    state = DATA_SENT;
}
void FlashProgrammer::handleDataSent()
{
    //data sent. Send request to erase and program the flash
    unlock( ERASE_AND_PROGRAMMING );
}

void FlashProgrammer::handleEraseAndProgramming()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError()) return;

    if( checkSdoAvailable(0x3300, 2)) return;
    int eraseAndProgramm = 4;
    int err = sdoClient->write( 0x3300,2,&eraseAndProgramm, CO_INT32 );
    if( err == 0 ) state =  DATA_PROGRAMMED;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x3300 subidx = 2, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleDataProgrammed()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;  
    state = CHECK_PROGRAMMING_STATUS;
}

void FlashProgrammer::handleCheckProgrammingStatus()
{
    int err;
    if( checkSdoAvailable(0x3300, 3)) return;
    err = sdoClient->read(0x3300,3,&controllerStatus, CO_INT32);
    if( err == 0 ) state = WAIT_FOR_STATUS;
    else
    {
        Lg << ERROR << "--Error submitting the read request for node: " << owner->getName() << " idx = 0x3300 subidx = 3, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleWaitForStatus()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    if( controllerStatus != 0x32 ) state = CHECK_PROGRAMMING_STATUS;
    else 
    {
        state = CALCULATE_PARITY;
        status = CHECKING_CONTROL_SUMS;
    }
}

void FlashProgrammer::handleCalculateParity()
{
    int err;
    if( checkSdoAvailable(0x3300, 2)) return;

    int req = 0x20;  // calculate parity
    err = sdoClient->write(0x3300,2, &req, CO_INT32 );
    if( err == 0 ) state = WAIT_FOR_CALCULATE_PARITY;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x3300 subidx = 3, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleWaitForCalculateParity()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    state = CHECK_RAM_PARITY;
}

void FlashProgrammer::handleCheckRamParity()
{
    int err;
    if( checkSdoAvailable(0x3301, 1)) return;
    err = sdoClient->read(0x3301,1,&ramParity, CO_INT32);
    if( err == 0 ) state = WAIT_FOR_RAM_PARITY;
    else
    {
        Lg << ERROR << "--Error submitting the read request for node: " << owner->getName() << " idx = 0x3301 subidx = 1, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleWaitForRamParity()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    Lg << INFO << "FlashProgrammer: RAM PARITY = 0x" << hex << ramParity << dec<< " for node: " << owner->getNodeId() << endl;
    state = CHECK_FOR_FLASH_PARITY;
}

void FlashProgrammer::handleCheckFlashParity()
{
    int err;
    if( checkSdoAvailable(0x3301, 2)) return;
    err = sdoClient->read(0x3301,2,&flashParity, CO_INT32);
    if( err == 0 ) state = WAIT_FOR_FLASH_PARITY;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x3301 subidx = 2, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleWaitForFlashParity()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    Lg << INFO << "FlashProgrammer: FLASH PARITY = 0x" << hex << flashParity << dec << " for node: " << owner->getNodeId() << endl;
    Lg << INFO << "FlashProgrammer: FILE PARITY = 0x" << hex << fileParity << dec<< " for node: " << owner->getNodeId() << endl;
    if( (flashParity == ramParity) /* && (ramParity == fileParity) */ )
    {
        state = UNLOCK_MAIN_MODE;
        status = BOOTING;
    }
    else
    {
        state = IDLE;
        Lg << ERROR << "--Error: wrong control sums! node left in programming mode." << endl;
    }
}

void FlashProgrammer::handleUnlockMainMode()
{    
    unlock(JUMP_TO_MAIN_MODE);
}

void FlashProgrammer::handleJumpToMainMode()
{    
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;

    owner->ignoreBootMessages(0);  // do not ignore BOOT-UP messages

    int goToMainMode = 0x01;
    sdoClient->write(0x3300,2,&goToMainMode, CO_INT32);
    Lg << INFO << "FlashProgrammer: reboot & reprogramming finish for node: " << owner->getNodeId() << endl;
    state = FINISHED;
    sdoClient->reset(); // there will be no response, since node will reboot
    autoResetTimer->start(3000);
}


int FlashProgrammer::unlock( int nextState)
{
    int err;
    if( checkSdoAvailable(0x3300, 1)) return ERROR_SDO_BUSY;
    int key = 0x12345678;
    err = sdoClient->write( 0x3300,1,&key, CO_INT32 );
    if( err == 0 ) state = nextState;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x3300, subidx = 1, error = " << err << endl;
        reset();
    }
    return err;
}

int FlashProgrammer::checkSdoAvailable(int idx, int subidx)
{
    if( !sdoClient->available() )
    {
        Lg << ERROR << "--Error: sdoChannel not available for node: " << owner->getName() << " idx = 0x" << hex <<idx <<", subidx = " << subidx << dec << endl;
        reset();
        return 1;
    }
    return 0;
}

int FlashProgrammer::checkSdoError()
{
    int err;
    int idx, sidx;
    if( (err = sdoClient->getError()) != NO_ERROR )
    {
        idx = sdoClient->getIndex();
        sidx = sdoClient->getSubindex();
        if( err == ERROR_ABORT ) Lg << ERROR << "--Error write(0x" << hex << idx << ", subidx = 0x" << sidx << dec << " for node: " << owner->getName() << ", ABORT. Code = " << sdoClient->getAbortDescription() << endl;
        else Lg << ERROR << "--Error sdo (0x" << hex << idx << ", subidx = 0x" << sidx << dec << " for node: " << owner->getName() << ", " << sdoClient->getErrorDescription() << endl;
        reset();
        sdoClient->reset();
        return 1;
    }
    sdoClient->reset();
    return 0;
}


int FlashProgrammer::unlockForErasingAndProgramming()
{
    status = REPROGRAMMING;
    return unlock( SINGLE_STEP_ERASE_AND_REPROGRAM );
}

void FlashProgrammer::handleSingleStepEraseAndReprogram()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError()) return;

    if( checkSdoAvailable(0x3300, 2)) return;
    int eraseAndProgramm = 4;
    int err = sdoClient->write( 0x3300,2,&eraseAndProgramm, CO_INT32 );
    if( err == 0 ) state =  SINGLE_STEP_DATA_PROGRAMMED;
    else
    {
        Lg << ERROR << "--Error submitting the write request for node: " << owner->getName() << " idx = 0x3300 subidx = 1, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleSingleStepDataProgrammed()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    state = SINGLE_STEP_CHECK_PROGRAMMING_STATUS;
}

void FlashProgrammer::handleSingleStepCheckProgrammingStatus()
{
    int err;
    if( checkSdoAvailable(0x3300, 3)) return;
    err = sdoClient->read(0x3300,3,&controllerStatus, CO_INT32);
    if( err == 0 ) state = SINGLE_STEP_WAIT_FOR_STATUS;
    else
    {
        Lg << ERROR << "--Error submitting the read request for node: " << owner->getName() << " idx = 0x3300 subidx = 3, error = " << err << endl;
        reset();
    }
}

void FlashProgrammer::handleSingleStepWaitForStatus()
{
    if( !sdoClient->done() ) return;
    if( checkSdoError() ) return;
    if( controllerStatus != 0x32 ) state = SINGLE_STEP_CHECK_PROGRAMMING_STATUS;
    else
    {
        state = IDLE;
        status = IDLE;
    }
}
