
#include "sdo.h"

#ifdef __BIG_ENDIAN_ORDERING__
    void littleEndian2bytes( void * data );   
    void littleEndian3bytes( void * data ); 
    void littleEndian4bytes( void * data );   
    void littleEndian5bytes( void * data );
    void littleEndian6bytes( void * data );   
    void littleEndian7bytes( void * data ); 
    void littleEndian8bytes( void * data );  

    void bytes2toLittleEndian( void * data, char * bytes );
    void bytes3toLittleEndian( void * data, char * bytes );
    void bytes4toLittleEndian( void * data, char * bytes );
    void bytes5toLittleEndian( void * data, char * bytes );
    void bytes6toLittleEndian( void * data, char * bytes );
    void bytes7toLittleEndian( void * data, char * bytes );
    void bytes8toLittleEndian( void * data, char * bytes );

    long bytesToLong( char * bytes ); 
#else
    void littleEndian2bytes( void * data ) {}   
    void littleEndian3bytes( void * data ) {}
    void littleEndian4bytes( void * data ) {}
    void littleEndian5bytes( void * data ) {} 
    void littleEndian6bytes( void * data ) {}   
    void littleEndian7bytes( void * data ) {}
    void littleEndian8bytes( void * data ) {}

    void bytes2toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 2 ); }
    void bytes3toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 3 ); } 
    void bytes4toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 4 ); }
    void bytes5toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 5 ); }
    void bytes6toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 6 ); }
    void bytes7toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 7 ); } 
    void bytes8toLittleEndian( void * data, char * bytes ) { memcpy( bytes, data, 8 ); }

    long bytesToLong( char * bytes )
    {
      return *((long*)bytes);   // the data in CANopen msg is packed as 'little endian'
    }
#endif


SDOclient::SDOclient(int id)
{
  construct( id, NULL );
}

SDOclient::SDOclient(int id, CanMessageCollector * collector )
{
  construct( id, collector );
}

void SDOclient::construct(int id, CanMessageCollector * collector)
{
  nodeID = id;
  receiveIdOffset = 0x580;   // default
  transmitIdOffset = 0x600;  // default
  dataStorageOffset = 0;
  localStorage = NULL;
  dataSize = -1;  // not indicated
  progress = 0;
  toggleBit = 0;
  
  cmc = collector;
  setName( "SDOclient" );

  state = IDLE;
  error = NO_ERROR;
  Timer = new HrTimer(SDO_TIMEOUT_MILISEC);
}
 
void SDOclient::activate()
{
  switch( state )
  {
  case IDLE: 
    break;
  case DONE:
    break;
  default:
    if( Timer->expired() ) 
      {
        error = ERROR_TIMEOUT;
        state = DONE;        
      }
    break; 
  }
}

void SDOclient:: requestNextSegment()
{
  // send the request for the next segment
  CanMessage m;
  char byte[8];

  toggleBit = !toggleBit;
  byte[0] =  SEGMENT_UPLOAD_REQUEST_CODE | (toggleBit << 4);
  memset( &byte[1], '\0', 7 );
  m.setId( nodeID + transmitIdOffset );  
  m.setData( byte );
  cmc->submit(m);
  Timer->restart();
  
  state = UPLOAD_SEGMENT;
  error = NO_ERROR;   
}

int SDOclient::isAbort(CanMessage& m)
{
 // check if ABORT SDO:

  unsigned char * data;
  int idx;

  data = m.getData();
  idx = ((int)data[2]) << 8 | data[1];

  if( ( data[0] & 0xF0 ) == ABORT_CODE )
    {
      if( (index == idx) && (subindex == data[3]) )
        {
          abortCode = ((((((int)data[7] << 8 ) | data[6] ) << 8 ) | data[5] ) << 8 )  | data[4];
          state = DONE;
          error = ERROR_ABORT;
       //   Lg << WARNING << "-Warning: node = 0x" << hex << nodeID << ", SDO(0x" << idx << ", 0x" << subindex << ") abort = 0x" << abortCode << dec <<", "<< getAbortDescription() << endl;
          return error;
        }
      else
        {
          // just a warning. If the expected SDO does not come, then ERROR_TIMOUT will occure 
          Lg <<  WARNING << "-Warning: node = 0x" << hex << nodeID << " ABORT message, but unexpected index, subindex ( 0x" << idx << ",0x" << (int)data[3] << " )" << dec << endl;         
        }
    }
  
  return 0;
}

void SDOclient::initiateDownload( CanMessage& m )
{
  unsigned char * data = m.getData();
  int idx = (((int)data[2]) << 8 ) | data[1];

  if( (data[0] & 0xF0) == INITIATE_DOWNLOAD_RESPONSE_CODE )
    {
      if( (index == idx) && (subindex == data[3]) )
        {
          if( dataStorageOffset == dataSize )
            {
              progress = 100;
              state = DONE; 
              error = NO_ERROR;
            }
          else 
            {
              toggleBit = 1;
              sendDataSegment();  // send first segment of segmented data  
              state = DOWNLOAD_SEGMENT;
              error = NO_ERROR;
            }
        }           
      else  Lg << WARNING << "-Warning: unexpected index, subindex for node = 0x" << hex << nodeID << ", index = 0x" << idx << ", subidx = 0x" << (int)data[3] << dec << endl; // just a warning. If the expected SDO does not come, then ERROR_TIMOUT will occure          
    }
  else
    {       
      if( (index == idx) && (subindex == data[3]) )
        {
          state = DONE;
          error = ERROR_BAD_SDO_SEQUENCE;
        }
      else Lg << WARNING << "Warning: unexpected (2) index, subindex for node = 0x" << hex << nodeID << ", index = 0x" << hex << idx << ", subidx = 0x" << data[3] <<  dec << endl;  // just a warning. If the expected SDO does not come, then ERROR_TIMOUT will occure         
    }
}

int SDOclient::sendDataSegment()
{
  CanMessage m;
  char bytes[8];

  int nBytes = dataSize - dataStorageOffset;

  if( nBytes == 0 ) return 0;  // all data sent

  if( nBytes > 7 ) nBytes = 7;

  memcpy( bytes + 1, dataStorage + dataStorageOffset, nBytes );
  dataStorageOffset += nBytes;
  toggleBit = !toggleBit;
  progress = (dataSize > -1) ? ((dataStorageOffset * 100) / dataSize) : 0 ;
  bytes[0] = SEGMENT_DOWNLOAD_REQUEST_CODE + ((7-nBytes)<< 1) + (dataStorageOffset == dataSize) + (toggleBit << 4);
  m.setId( nodeID + transmitIdOffset );  
  m.setData( bytes );
  cmc->submit(m);
  Timer->restart();
  return 1;  // not all data sent yet
}

void SDOclient::downloadSegment( CanMessage& m )
{
  unsigned char * data = m.getData();

  if( (data[0] & 0xE0) == SEGMENT_DOWNLOAD_RESPONSE_CODE )
    {                
      // check the toggle bit:
      if( (unsigned int)((data[0] & 0x10) >> 4 ) != toggleBit )
        {
          state = DONE;
          error = ERROR_BAD_TOGGLE_BIT;                
        }
      else
        {     
          if( !sendDataSegment() )
            {
              state = DONE;
              error = NO_ERROR;
            }
        }        
    }
  else    
    {
      state = DONE;
      error = ERROR_BAD_SDO_SEQUENCE;
    } 
}

void SDOclient::initiateUpload( CanMessage& m )
{
  unsigned char * data = m.getData();
  int idx = ((int)data[2]) << 8 | data[1];

  if( (data[0] & 0xF0) == INITIATE_UPLOAD_CODE )
    {
      if( (index == idx) && (subindex == data[3]) )
        {
          if( data[0] & 0x2 )
            {
              if( data[0] & 0x1 ) dataSize = 4 - ((data[0] & 0x0C) >> 2 ); // expedited transfer and size indicated
              else dataSize = 4;  // size not indicated, but programm has to guess some size
              memcpy( dataStorage, data+4, dataSize );               
              convert();
              progress = 100;
              state = DONE; 
              error = NO_ERROR;
            }
          else
            {
              if( data[0] & 0x1 )  dataSize = bytesToLong( (char*)(data + 4) );        // read data length, if indicated            
              else                 dataSize = -1;               
              toggleBit = 1;  // requestNextSegment toggles the bit before sending a message, the first msg in sequence should have toggle bit = 0
              requestNextSegment();                           
            }
        }
      else  Lg << WARNING << "-Warning: unexpected index, subindex for node = 0x" << hex << nodeID << ", idx = 0x" << idx << ", subidx = 0x" << data[3] << dec << endl; // just a warning. If the expected SDO does not come, then ERROR_TIMOUT will occure          
    }
  else
    {       
      if( (index == idx) && (subindex == data[3]) )
        {
          state = DONE;
          error = ERROR_BAD_SDO_SEQUENCE;
        }
      else Lg << WARNING << "-Warning: unexpected index, subindex for node = 0x" << hex << nodeID << ", idx = 0x" << idx << ", subidx = 0x" << data[3] << dec << endl;  // just a warning. If the expected SDO does not come, then ERROR_TIMOUT will occure         
    }
}

void SDOclient::uploadSegment( CanMessage& m )
{
  unsigned int size;
  unsigned char * data = m.getData();

  if( (data[0] & 0xE0) == SEGMENT_UPLOAD_RESPONSE_CODE )
    {                
      // check the toggle bit:
      if( (unsigned int)((data[0] & 0x10) >> 4 ) != toggleBit )
        {
          state = DONE;
          error = ERROR_BAD_TOGGLE_BIT;                
        }
      else
        {            
          // fetch data from the message
          size = 7 - ((data[0] & 0x0E) >> 1 );
          memcpy( dataStorage + dataStorageOffset, data+1, size );                  
          dataStorageOffset += size;        
          convert();               
          if( data[0] & 0x01 )  // completion bit
            {
              progress = 100;
              state = DONE;
              error = NO_ERROR;
            }
          else
            {
              progress = (dataSize > -1) ? ((dataStorageOffset * 100) / dataSize) : 0 ;
              requestNextSegment(); 
            }
        }        
    }
  else    
    {
      state = DONE;
      error = ERROR_BAD_SDO_SEQUENCE;
    }
}

int SDOclient::handle( CanMessage& m )
{
  
  if( isAbort(m) ) return 0; 
 
  switch( state )
  {
  case IDLE: 
  case DONE:
    Lg << WARNING << "-Warning: SDO unexpected message arrived" << endl;
    m.dump();
    return -1;
    break;

  case INITIATE_UPLOAD:
    initiateUpload(m);
    return 0;
    break;

  case UPLOAD_SEGMENT:
    uploadSegment(m);
    return 0;
    break;

  case INITIATE_DOWNLOAD:
    initiateDownload(m);
    return 0;
    break;

  case DOWNLOAD_SEGMENT:
    downloadSegment(m);
    return 0;
    break;

  default:
    Lg << ERROR <<  "--Error: SDOclient::handle() - unknown state:" << state << endl;
    return -1;
    break;
  }

}

int SDOclient::read( int idx, int subidx, int dSize )
{
  if(( localStorage = (char*)malloc( dSize )) == NULL ) return ERROR_NOT_ENOUGH_MEMORY;
  memset( localStorage, '\0', dSize );
  return read( idx, subidx,  localStorage, CO_DATA_TYPE_UNKNOWN );
}
int SDOclient::read( int idx, int subidx, void * data, int dType )
{
  CanMessage m;
  char byte[8];

  if( state != IDLE ) return ERROR_SDO_BUSY;

  dataStorageOffset = 0;
  dataSize = -1;  // not indicated
  progress = 0;
 
  byte[0] = INITIATE_UPLOAD_CODE;
  byte[1] = (unsigned char )(idx & 0x00FF);
  byte[2] = (unsigned char )(idx >> 8);
  byte[3] = (unsigned char )subidx;
  byte[4] = byte[5] = byte[6] = byte[7] = 0;
  m.setId( nodeID + transmitIdOffset );  
  m.setData( byte );
  cmc->submit(m);
  Timer->restart();
  
  // store request parameters
  index = idx;
  subindex =  subidx;
  dataStorage = (char*)data;
  dataType = dType;
  access = READ; 
  state = INITIATE_UPLOAD;

  return NO_ERROR; 
}

int SDOclient::write( int idx, int subidx, void * data, int dType )
{

  CanMessage m;
  char byte[8];

  if( state != IDLE ) return ERROR_SDO_BUSY;
 
  byte[0] = INITIATE_DOWNLOAD_CODE;
  byte[1] = (unsigned char )(idx & 0x00FF);
  byte[2] = (unsigned char )(idx >> 8);
  byte[3] = (unsigned char )subidx;
  
  switch( dType )
    {
    case CO_BOOL1:
    case CO_INT8:
    case CO_UINT8:
      byte[0] |= ( 3 + 12 );
      byte[4] = *((char*)data);
      byte[5] = byte[6] = byte[7] = 0;
      dataStorageOffset = 1;
      dataSize = 1;
      break;
    case CO_INT16:
    case CO_UINT16:      
      byte[0] |= ( 3 + 8 );
      bytes2toLittleEndian( data, &byte[4] );
      byte[6] = byte[7] =  0;        
      dataStorageOffset = 2;
      dataSize = 2;
      break;
    case CO_INT24:
    case CO_UINT24: 
      byte[0] |= ( 3 + 4 );
      bytes3toLittleEndian( data, &byte[4] );
      byte[7] =  0;
      dataStorageOffset = 3;
      dataSize = 3;
      break;
    case CO_INT32:
    case CO_UINT32:
    case CO_REAL32:   
      byte[0] |= ( 3 );
      bytes4toLittleEndian( data, &byte[4] );  
      dataStorageOffset = 4;
      dataSize = 4;
      break;
    case CO_INT40:
    case CO_UINT40: 
      byte[0] |= ( 1 );
      dataStorageOffset = 0;
      dataSize = 5;  
      bytes4toLittleEndian( &dataSize, &byte[4] );       
      bytes5toLittleEndian( (char*)data, (char*)data );     
      break;
    case CO_INT48:
    case CO_UINT48:      
      byte[0] |= ( 1 );  
      dataStorageOffset = 0;
      dataSize = 6;
      bytes4toLittleEndian( &dataSize, &byte[4] );      
      bytes6toLittleEndian( (char*)data, (char*)data );       
      break;

    case CO_INT56:
    case CO_UINT56: 
      byte[0] |= ( 1 );   
      dataStorageOffset = 0;
      dataSize = 7;     
      bytes4toLittleEndian( &dataSize, &byte[4] );   
      bytes7toLittleEndian((char*)data,(char*)data );
      break;

    case CO_REAL64:
    case CO_INT64:      
    case CO_UINT64:  
      byte[0] |= ( 1 );  
      dataStorageOffset = 0;
      dataSize = 8;
      bytes4toLittleEndian( &dataSize, &byte[4] );   
      bytes8toLittleEndian( (char*)data, (char*)data );
      break;

    case CO_TIME_OF_DAY:
    case CO_TIME_DIFF:
      dataStorageOffset = 0;
      dataSize = 6;
      byte[0] |= ( 1 );     
      bytes4toLittleEndian( &dataSize, &byte[4] );       
      bytes6toLittleEndian((char*)data, (char*)data);  //warning: struct. Maybe another packing method?
      break;
    
    case CO_VIS_STR: 
      dataStorageOffset = 0;
      dataSize = strlen( (char*)data ); 
      byte[0] |= ( 1 );   // size indicated
      bytes4toLittleEndian( &dataSize, &byte[4] );   
      break;

    case CO_UNICODE_STR:
      dataStorageOffset = 0;
      dataSize = 2 * strlen( (char*)data ); 
      byte[0] |= ( 1 );
      bytes4toLittleEndian( &dataSize, &byte[4] );   
      break;

    case CO_DATA_TYPE_UNKNOWN:
    case CO_BIT_STR:
    case CO_OCTET_STR:
    case CO_DOMAIN:
      // size must be set by the setDataSize() function
      dataStorageOffset = 0;  
      if( dataSize < 5 )
        {
          byte[0] |= ( 3 + ((4-dataSize) << 2) );
          for( int i = 0; i <dataSize; i++ ) byte[4+i] = *(((char*)data) + i);
          dataStorageOffset = dataSize;   
        }
      else 
        {
          byte[0] |= ( 1 );
          bytes4toLittleEndian( &dataSize, &byte[4] );   
          dataStorageOffset = 0;   
        }
      break;
    }

  m.setId( nodeID + transmitIdOffset );  
  m.setData( byte );
  cmc->submit(m);
  
  // store request parameters
  index = idx;
  subindex =  subidx;
  dataStorage = (char*)data;
  dataType = dType;
  access = WRITE; 
  state = INITIATE_DOWNLOAD;
  Timer->restart();

  return NO_ERROR; 

}

void SDOclient::convert()
{
  switch( dataType )
  {    
    case CO_BOOL1:
    case CO_INT8:
    case CO_UINT8:
       break;    // no convertion is needed: the first byte represents the value

    case CO_INT16: 
    case CO_UINT16:
      littleEndian2bytes( dataStorage );
      break;
    
  case CO_INT32:
  case CO_UINT32:    
  case CO_REAL32: 
    littleEndian4bytes( dataStorage );
    break;
  
  case CO_REAL64:  
  case CO_INT64: 
  case CO_UINT64:     
    littleEndian8bytes( dataStorage );
    break;
  
  case CO_INT40:    
  case CO_UINT40: 
    littleEndian5bytes( dataStorage );
    break;
  
  case CO_INT48:  
  case CO_UINT48:  
    littleEndian6bytes( dataStorage );  
    break;
  
  case CO_INT56:
  case CO_UINT56: 
    littleEndian7bytes( dataStorage );  
    break;

  case CO_UINT24:  
  case CO_INT24:
    littleEndian3bytes( dataStorage );
    break;
  
  case CO_VIS_STR: // no conversion needed
  case CO_OCTET_STR: 
  case CO_BIT_STR:
  case CO_UNICODE_STR:
  case CO_TIME_OF_DAY:
  case CO_TIME_DIFF:
  case CO_DOMAIN:    
  case CO_DATA_TYPE_UNKNOWN:
    break;

  default:
    Lg << WARNING << "-Warning: unexpected data type code of node 0x " << hex << nodeID << dec << dataType << ", ( 0x" << hex << dataType << ")" << dec << endl;
    break;    
  }
  
}

char * ErrorText[] =
  {
    "no error",
    "timeout",
    "bad sequence",
    "bad toggle bit",
    "abort",
    "busy",
    "unhandled SDO type"
  } ;


 char * SDOclient::getAbortDescription()
{
  return getAbortDescription( abortCode );
}
 char * SDOclient::getAbortDescription(int errCode)
{
  switch( errCode )
    {
    case 0 : return "no error";
    case 0x5040000: return "SDO protocol timed out";
    case 0x5030000: return"Toggle bit not alternated";
    case 0x5040001: return "Invalid SDO command specifier";
    case 0x5040002: return "Invalid block size (block mode)";
    case 0x5040003: return "Invalid sequence number(block mode)";
    case 0x5040004: return "CRC error (block mode)";
    case 0x5040005: return "Out of memory";
    case 0x6010000: return "Unsupported access to an object";
    case 0x6010001: return "Attempt to read a write-only object";
    case 0x6010002: return "Attempt to write a read only object";
    case 0x6020000: return "Object does not exist";
    case 0x6040041: return "Object can not be mapped to the PDO";
    case 0x6040042: return "The number and length of the objects to be mapped would exceed PDO length.";
    case 0x6020043: return "General parameter incompatibility";
    case 0x6060000: return "Access failed due to hardware error";
    case 0x6070010: return "Data type does not match. Length of service parameter does not match";
    case 0x6070012: return "Data type does not match. Length of service parameter is too high";
    case 0x6070013: return "Data type does not match. Length of service parameter is too short";
    case 0x6090011: return "Subindex does not exist";
    case 0x6090030: return "Value range of parameter exceeded";
    case 0x6090031: return "Value range of parameter written is too high";
    case 0x6090032: return "Value range of parameter written is too low";
    case 0x6090036: return "Maximum value is less than the minimum value";
    case 0x8000000: return "General error";
    case 0x8000020: return "Data can not be transfered or stored to the application";
    case 0x8000021: return "Data can not be transfered or stored to the application because of local control";
    case 0x8000022: return "Data can not be transfered or stored to the application because of the present device state";
    case 0x8000023: return "Object Dictionary dynamic generation failed or no Object Dictionary present";
    }
  return "unknown error code";
}

 char *SDOclient::getErrorDescription( unsigned int errCode )
{
  if( errCode < sizeof(ErrorText)/sizeof(ErrorText[0]))  return ErrorText[errCode];

  return "Undefined error";
}

 char *SDOclient::getErrorDescription()
{
  return getErrorDescription(error);
}
