

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <signal.h>


#include "shmlogger.h"


#define DEFAULT_PATH "/var"
#define MAX_FILES_PRO_DAY 3
#define MAX_FILE_SIZE 1000000L

static void  cleanup( void );
static void  saveMsg( LOGMSG *msg );
static void  checkFile( void );
static int   createLogFile(long date);
static long  getDate( void );
static void  parseArgs(int argc, char * argv[]);
static void  help( void );
static void  terminate( int signo );

static char  Format[64];
static int   ConsoleMode = 0;
static char  Path[512];
static int   MaxFilesProDay = MAX_FILES_PRO_DAY;
static long  MaxFileSize = MAX_FILE_SIZE;
static int   FileCounter = 0;
static long  CurrentDate;
static FILE* LogFileHandle = NULL;
static char  LogFileName[256];

int main( int argc, char * argv[] )
{
  int errCode;
  LOGMSG  msg;

  parseArgs( argc, argv );

  signal( SIGINT,  terminate );
  signal( SIGQUIT, terminate );
  signal( SIGABRT, terminate );
  signal( SIGTERM, terminate );
  signal( SIGTSTP, terminate );

  atexit( cleanup );
  
  sprintf( Format, "%s%d%s", "%02d.%02d.%02d %02d:%02d:%02d %", MAX_PROCESS_NAME, "s [%d]: %s\n");
  
  if(ConsoleMode) printf( "logman vers. %s, %s started.\n", __DATE__, __TIME__ );

  if( (errCode = initLog("logman")) != 0 )
    {
      if(ConsoleMode) printf( "openLog() error = %d, %s\n", errCode, strerror(errCode));
      exit(errCode);
    }

  while( 1 )
    {
      while( getLog( &msg ) ) saveMsg( &msg );
      usleep(300000L);
    }
  
}

void terminate( int signo )
{
    char buff[32];
    
    sprintf( buff, "signal %d cought", signo );
    putLog( buff );
    exit(0);
}


void parseArgs(int argc, char * argv[])
{
  int i;

  strcpy(Path, DEFAULT_PATH );
  MaxFilesProDay = MAX_FILES_PRO_DAY;
  MaxFileSize    = MAX_FILE_SIZE;
  
  for( i = 1; i < argc; i++ )
    {
      if( !strcmp( argv[i], "-d" ))
	{
	  ConsoleMode = 1;
	  continue;
	}

      if( !strcmp( argv[i], "-p" ))
	{
	  if( i + i < argc ) strcpy( Path, argv[i+1] );
	  else { puts(" missing path name" ); exit(1); }
	  continue;
	}
       if( !strcmp( argv[i], "-m" ))
	{
	  if( i + i < argc )  MaxFilesProDay = atoi( argv[i+1] );
	  else { puts(" missing max number of files" ); exit(1); }
	  continue;
	}
       if( !strcmp( argv[i], "-s" ))
	{
	  if( i + i < argc )  MaxFileSize = atoi( argv[i+1] );
	  else { puts(" missing value of max size" ); exit(1); }
	  continue;
	}
       help();	
    }
}

void cleanup( void )
{
    char buff[32];

    sprintf( buff, "terminated. Exit code = %d", errno );
    putLog( buff );
    closeLog();
}

void saveMsg( LOGMSG * msg )
{
  int errCode;
  char buff[60+MAX_MESSAGE + MAX_PROCESS_NAME];
  struct tm * t  = localtime( &( msg->timeStamp ) );

  sprintf( buff, Format , t->tm_mday, t->tm_mon+1, t->tm_year%100, t->tm_hour, t->tm_min, t->tm_sec,
	   msg->pName, msg->pid, msg->message);

   if(ConsoleMode) printf( buff );

  checkFile();
  
  if( LogFileHandle ==NULL) return;
  
  clearerr( LogFileHandle );
  
  fwrite( buff, sizeof(char), strlen(buff), LogFileHandle );
  if( (errCode = ferror( LogFileHandle ))) 
    {
      if(ConsoleMode) printf("logman: can not write to output file: %s, error = %d, %s\n", LogFileName, errCode, strerror(errCode));
    }
  else
    fflush( LogFileHandle );
  
}



void checkFile( void )
{
  /* check, if (1) file is not bigger than MAX_FILE_SIZE or (2) if date is changed. If yes, then close
   log file and open a new one. */
  long d;
  
  if( LogFileHandle == NULL )
    {
      /* log file is not opened yet, create a new one. */
      createLogFile( getDate() );
      CurrentDate = getDate();
      return;
    }

  if( ((d = getDate()) != CurrentDate ) || ftell( LogFileHandle ) > MaxFileSize ) 
    {
      if( LogFileHandle ) fclose( LogFileHandle );
      CurrentDate = d;
      FileCounter = 0;
      createLogFile( d );
      return;
    }
}

int createLogFile(long date)
{
  if( FileCounter == MAX_FILES_PRO_DAY ) FileCounter = 0;
  
  if( FileCounter++ ) sprintf( LogFileName, "%s/slog%06ld_%d.log", Path,  date, FileCounter ); 
    else sprintf( LogFileName, "%s/slog%06ld.log", Path, date );
  LogFileHandle = fopen( LogFileName, "a" );
  
  return LogFileHandle == NULL ? errno : 0;    
}


long getDate( void )
{
  /* returns YYMMDD */
  struct tm * t;
  time_t timer = time(NULL);
  t = localtime( &timer );

  return  (t->tm_year%100)*10000L + (t->tm_mon+1)*100 + t->tm_mday;
}


void help( void )
{
    printf( "\n\nlogman vers. %s, %s\n", __DATE__, __TIME__ );
    printf( "___________________________________\n" );
    printf( "available options are:\n" );
    printf( "-d          print messages on console\n" );
    printf( "-p <path>   store log files in <path> directory\n" );
    printf( "-m <number> store max <number> log files pro day\n" );
    printf( "-s <size>   allow log files not bigger than <size>\n" );
    printf( "\n\n" );

    exit(0);
}
