
/*
        FILE:     shmlogger.c
	PURPOSE:  implementation of shared memory based message logging system.
	          Access to the shared memory is __NOT__ semaphore-protected.
	HISTORY:
	          19.05.2003, Piotr Bartkiewicz:   creation of version 1.0
		  08.01.2008, Piotr Bartkiewicz:   setProcessName( char * processName ) added.
			      	
 */


#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>

#include "shmlogger.h"

#define N_LOGMSG 500
#define SHM_NAME "/local/home/fecadmin"
#define SHM_KEY  0x12031964     /* arbitrary choosen, if ftok() fails. See openLog() */
#define SHM_MODE ((SHM_R | SHM_W) | (( SHM_R >> 3 ) | (SHM_W >> 3 )) | (( SHM_R >> 6 ) | (SHM_W >> 6 )))


typedef struct
{
    long readPtr;
    long writePtr;
} LOGHDR;


typedef struct
{
    LOGHDR hdr;
    LOGMSG msg[N_LOGMSG];
} LOGSHM;


static LOGSHM * LogShm;
static char ProcessName[MAX_PROCESS_NAME+1];
static char TerminalMode = 0;

#include <stdio.h>

void setProcessName( char * processName )
{
  strncpy( ProcessName, processName, MAX_PROCESS_NAME );
  ProcessName[MAX_PROCESS_NAME] = '\0';
}

int openLog( char * processName )
/*
    Attaches/creates the shared memory block, initializes the pointer in case of creation.
    Returns 0, if no error, otherwise error number
*/
{
#ifndef vxworks    /*----------------[UNIX part]-----------------------------------*/

    int key, id;
    struct shmid_ds id_ds;
    
    /* try to obtain shared memory key */
    if( (key  = ftok( SHM_NAME, 1) ) < 0 ) key = SHM_KEY;

    /* check, if the shared memory already exists */
    if( (id = shmget(key, sizeof(LOGSHM), SHM_MODE | IPC_CREAT | IPC_EXCL )) < 0 )
      {
	if( errno != EEXIST ) return errno;
          
	/* try to obtain  shared memory ID. If shm doesn't exist, it will be created. */    
	if( (id = shmget(key, sizeof(LOGSHM), SHM_MODE )) < 0 ) return errno; 
      }
    else
      {
	/* set read-write rights for all */
	if( shmctl( id, IPC_STAT, &id_ds ) < 0 ) return errno;
	id_ds.shm_perm.mode =  SHM_MODE;
	if( shmctl( id, IPC_SET, &id_ds ) < 0 ) return errno;
      }

    /* attach to shared memory */
    if( (LogShm = (LOGSHM*)shmat(id, 0, 0)) == (LOGSHM *) -1 ) return errno;
     

#else  /*---------------[vxWorks part]---------------------------------------------*/
      
    if( (LogShm = (LOGSHM*)calloc( sizeof(LOGSHM), sizeof(char)) == (LOGSHM*) NULL ) return errno;
#endif
    
    setProcessName( processName );
    return 0;	
}

void vPutLog( char * fmt, ... )
{
    va_list list;
    char *p, *string, *msg, *buff;
    int integer;
    double dbl;
    char c;
    char tmpbuff[128], tmpmsg[128];

    buff = tmpbuff;
    msg = tmpmsg;
    memset( tmpmsg, '\0', sizeof( tmpmsg ));
    va_start( list, fmt );
    for( p = fmt; *p; ++p )
    {
	if( *p != '%' ) *msg++ = *p;
	else
	{
	    switch( *++p )
	    {
	     case 's':
		 string = va_arg( list, char * );
		 while( *string ) *msg++ = *string++; 	       		 
		 break;
	     case 'd':
		 integer = va_arg( list, int );
		 sprintf( buff, "%d", integer );
		 while( *buff ) *msg++ = *buff++ ;
		 break;
	     case 'f':
		 dbl = va_arg( list, double );
		 sprintf( buff, "%f", dbl );
		 while( *buff ) *msg++ = *buff++ ;
		 break;
	     case 'c':
               /* c = va_arg( list, char ); */
                 c = va_arg( list, int );  
                 /* see compiler warning:shmlogger.c:134: warning: `char' is promoted to `int' when passed through `...'
                    shmlogger.c:134: warning: (so you should pass `int' not `char' to `va_arg')
                    shmlogger.c:134: note: if this code is reached, the program will abort
                 */   
		 sprintf( buff, "%c", c );
		 while( *buff ) *msg++ = *buff++ ;
		 break;
	     default:
		 *msg++ = *p;
	    }	    
	}	
    }
    va_end( list );
    putLog( tmpmsg );
}
    
    
    
void putLog( const char * msg )
{
    LOGMSG m;

    m.timeStamp = time(NULL);
    m.pid = getpid();
    strcpy( m.pName, ProcessName );
    strncpy( m.message, msg, MAX_MESSAGE );
    m.message[MAX_MESSAGE] = '\0';

    /* lines below are critical section, which should be protected by a semaphore
       to ensure, that no message is lost (overwritten). But it isn't. */
    if(++(LogShm->hdr.writePtr) >= N_LOGMSG) LogShm->hdr.writePtr = 0;
    memcpy( &(LogShm->msg[LogShm->hdr.writePtr]), &m, sizeof(LOGMSG));
    if( TerminalMode ) puts( m.message );
}


int getLog( LOGMSG *m )
{
    long rdPtr =  LogShm->hdr.readPtr;
    
    if( LogShm->hdr.writePtr == rdPtr ) return 0;

    if( ++rdPtr >= N_LOGMSG) rdPtr = 0;
    memcpy( m, &(LogShm->msg[rdPtr]), sizeof(LOGMSG));
    LogShm->hdr.readPtr = rdPtr;
    return 1;
}

int getLogInfo( long * lastWritePtr,  long * bufferSize )
{
  /* return 1 if  *lastWritePtr points to last messages in the shared memory buffer */
  long tmp = *lastWritePtr;

  *lastWritePtr = LogShm->hdr.writePtr;
  *bufferSize   = N_LOGMSG;
  
  return ( LogShm->hdr.writePtr == tmp );
}

int readLog( long msgNo, LOGMSG * msg )
{
  /* copy message pointed by msgNo to * msg and return number of messages available in the shared memory buffer */
  
  memcpy( msg, &(LogShm->msg[msgNo]), sizeof( LOGMSG ));
  if( msgNo > LogShm->hdr.writePtr ) return N_LOGMSG - msgNo + LogShm->hdr.writePtr;
  return LogShm->hdr.writePtr - msgNo;
}

int initLog( char * processName )
{
/* this function should be called only by logman process or other compatible log manager */
    int errCode;

    if( (errCode = openLog( processName )) != 0 ) return errCode;
    memset( &(LogShm->hdr), '\0', sizeof( LOGHDR ));

    return 0;
}

int closeLog( void )
{
    return shmdt( (char*) LogShm );
}

char terminalMode( char t )
{
   char prev = TerminalMode;
   TerminalMode = t;
   return prev;
}

